import numpy as np

import libphoebe
from phoebe.backend.mesh import euler_trans_matrix, transform_position_array
from phoebe.frontend import io

try:
    import phoebe_legacy as phb1
except ImportError:
    try:
        import phoebeBackend as phb1
    except ImportError:
        _can_phb = False
    else:
        _can_phb = True
else:
    _can_phb = True
    _phb_init = False

import logging

logger = logging.getLogger("HORIZON_ANALYTIC")


def cartesian_to_polar(x, y, x0=0.0, y0=0.0):

    rhos = np.sqrt((x-x0)**2 + (y-y0)**2)
    thetas = np.arctan2(y-y0, x-x0)

    return rhos, thetas

def polar_to_cartesian(rho, theta, x0=0.0, y0=0.0):
    x = rho * np.cos(theta) + x0
    y = rho * np.sin(theta) + y0

    return x, y

def marching(q, F, d, Phi, scale, euler, pos):
    """

    viewing_vector is in ROCHE coordinate system so not necessarily along z
    q, F, d, Phi should be body._mesh_args
    scale should be body._scale
    """

    euler = np.asarray(euler)
    pos = np.asarray(pos)

    # print "*** horizon_analytic.marching pos", pos
    # print "*** horizon_analytic.marching euler", euler

    viewing_vector = transform_position_array(array=np.array([0, 0, -1]),
                                              pos=np.array([0., 0., 0.]),
                                              euler=euler,
                                              is_normal=True,
                                              reverse=True)

    # print "*** horizon_analytic.marching viewing vector", viewing_vector

    # note: all of this is done in the ROCHE coordinate system.
    horizon_roche = libphoebe.roche_horizon(viewing_vector, q, F, d, Phi, choice=0)

    horizon_roche *= scale

    horizon_orbit = transform_position_array(array=horizon_roche,
                                             pos=pos,
                                             euler=euler,
                                             is_normal=False,
                                             reverse=False)

    horizon_rhos, horizon_thetas = cartesian_to_polar(horizon_orbit[:,0], horizon_orbit[:,1], pos[0], pos[1])

    return {'xs': horizon_orbit[:,0], 'ys': horizon_orbit[:,1], 'zs': horizon_orbit[:,2],
            'rhos': horizon_rhos, 'thetas': horizon_thetas}


def wd(b, time, scale, pos):



    def rho(theta, c, s):
        sum = 0.0
        for i in range(len(c)):
            sum += c[i]*np.cos(i*theta)+s[i]*np.sin(i*theta)
        return sum

    if not _can_phb:
        return {'xs': [], 'ys': [], 'zs': [],
                'rhos': [], 'zs': []}

    if not _phb_init:
        phb.init()
        if hasattr(phb, 'auto_configure'):
            # then phb is phoebe_legacy
            phb.auto_configure()
        else:
            # then phb is phoebeBackend
            phb.configure()


    # TODO: move this outside the loop into backends.py?
    # TODO: make this work without the need to add an LC dataset
    # io.pass_to_legacy(b, 'change_to_tmpfile')
    legacy_dict = io.pass_to_legacy(b, compute=compute, **kwargs)
    io.import_to_legacy(legacy_dict)
    #phb.open('change_to_tmpfile')
    phb.setpar('phoebe_indep', 'Time (HJD)')
    combo = phb.lc((time,), 0, 0, 1)
    # phb.quit()

    if not len(combo[1]['theta']):
        return {'xs': [], 'ys': [], 'zs': [],
                'rhos': [], 'thetas': []}

    thetas = np.linspace(-np.pi, np.pi, 1001)
    rhos = rho(thetas, combo[1]['hAc'], combo[1]['hAs'])

    rhos *= scale

    xs, ys = polar_to_cartesian(rhos, thetas, pos[0], pos[1])

    return {'xs': xs, 'ys': ys, 'zs': np.ones(xs.shape)*pos[2],
            'rhos': rhos, 'thetas': thetas}



import numpy as np
from copy import deepcopy as _deepcopy

try:
    import dynesty as _dynesty
except ImportError:
    _use_dynesty = False
else:
    _use_dynesty = True

try:
    import emcee
except ImportError:
    _use_emcee = False
else:
    _use_emcee = True

_skip_filter_checks = {'check_default': False, 'check_visible': False}

def process_mcmc_chains_from_solution(b, solution, burnin=None, thin=None, lnprob_cutoff=None, adopt_parameters=None, flatten=True):
    """
    Process the full MCMC chain and expose the `lnprobabilities` and `samples`
    after applying `burnin`, `thin`, `lnprob_cutoff`.

    See also:
    * <phoebe.helpers.process_mcmc_chains>
    * <phoebe.parameters.solver.sampler.emcee>

    Arguments
    ---------------
    * `b` (<phoebe.frontend.bundle.Bundle>): the Bundle
    * `solution` (string): solution label
    * `burnin` (int, optional, default=None): If not None, will override
        the value in the solution.
    * `thin` (int, optional, default=None): If not None, will override
        the value in the solution.
    * `lnprob_cutoff` (float, optional, default=None): If not None, will override
        the value in the solution.
    * `adopt_parameters` (list, optional, default=None): If not None, will
        override the value in the solution.
    * `flatten` (bool, optional, default=True): whether to flatten to remove
        the "walkers" dimension.  If False, `lnprob_cutoff` will replace entries
        with nans rather than excluding them.

    Returns
    -----------
    * (array, array): `lnprobablities` (flattened to 1D if `flatten`) and `samples`
        (flattened to 2D if `flatten`) after processing
    """
    solution_ps = b.get_solution(solution=solution, **_skip_filter_checks)
    adopt_inds, adopt_uniqueids = b._get_adopt_inds_uniqueids(solution_ps, adopt_parameters=adopt_parameters)

    return process_mcmc_chains(solution_ps.get_value(qualifier='lnprobabilities', **_skip_filter_checks),
                               solution_ps.get_value(qualifier='samples', **_skip_filter_checks),
                               solution_ps.get_value(qualifier='burnin', burnin=burnin, **_skip_filter_checks),
                               solution_ps.get_value(qualifier='thin', thin=thin, **_skip_filter_checks),
                               solution_ps.get_value(qualifier='lnprob_cutoff', lnprob_cutoff=lnprob_cutoff, **_skip_filter_checks),
                               adopt_inds=adopt_inds, flatten=flatten)

def process_mcmc_chains(lnprobabilities, samples, burnin=0, thin=1, lnprob_cutoff=-np.inf, adopt_inds=None, flatten=True):
    """
    Process the full MCMC chain and expose the `lnprobabilities` and `samples`
    after applying `burnin`, `thin`, `lnprob_cutoff`.

    See also:
    * <phoebe.helpers.process_mcmc_chains_from_solution>

    Arguments
    -------------
    * `lnprobabilities` (array): array of all lnprobabilites as returned by MCMC.
        Should have shape: (`niters`, `nwalkers`).
    * `samples` (array): array of all samples from the chains as returned by MCMC.
        Should have shape: (`niters`, `nwalkers`, `nparams`).
    * `burnin` (int, optional, default=0): number of iterations to exclude from
        the beginning of the chains.  Must be between `0` and `niters`.
    * `thin` (int, optional, default=1): after applying `burnin`, take every
        `thin` iterations.  Must be between `1` and (`niters`-`burnin`)
    * `lnprob_cutoff` (float, optiona, default=-inf): after applying `burnin`
        and `thin`, reject any entries (in both `lnprobabilities` and `samples`)
        in which `lnprobabilities` is below `lnprob_cutoff`.
    * `adopt_inds` (list, optional, default=None): if not None, only expose
        the parameters in `samples` with `adopt_inds` indices.  Each entry
        in the list should be between `0` and `nparams`.
    * `flatten` (bool, optional, default=True): whether to flatten to remove
        the "walkers" dimension.  If False, `lnprob_cutoff` will replace entries
        with nans rather than excluding them.

    Returns
    -----------
    * (array, array): `lnprobablities` (flattened to 1D if `flatten`) and `samples`
        (flattened to 2D if `flatten`) after processing

    Raises
    ------------
    * TypeError: if `burnin` or `thin` are not integers
    * ValueError: if `lnprobabilities` or `samples` have the wrong shape
    * ValueError: if `burnin`, `thin`, or `adopt_inds` are not valid given
        the shapes of `lnprobabilities` and `samples`.
    """
    if not isinstance(burnin, int):
        raise TypeError("burnin must be of type int")
    if not isinstance(thin, int):
        raise TypeError("thin must be of type int")

    if len(lnprobabilities.shape) != 2:
        raise ValueError("lnprobablities must have shape (niters, nwalkers), not {}".format(lnprobabilities.shape))
    if len(samples.shape) != 3:
        raise ValueError("samples must have shape (niters, nwalkers, nparams), not {}".format(samples.shape))

    if lnprobabilities.shape[0] != samples.shape[0] or lnprobabilities.shape[1] != samples.shape[1]:
        raise ValueError("lnprobablities and samples must have same size for first 2 dimensions lnprobabilities.shape=(niters, nwalkers), samples.shape=(niters, nwalkers, nparams)")

    niters, nwalkers, nparams = samples.shape

    if adopt_inds is not None:
        if not isinstance(adopt_inds, list):
            raise TypeError("adopt_inds must be of type list or None")
        if not np.all([ai < nparams and ai >=0 for ai in adopt_inds]):
            raise ValueError("each item in adopt_inds must be between 0 and {}".format(nparams))

    if burnin < 0:
        raise ValueError("burnin must be >=0")
    if burnin >= niters:
        raise ValueError("burnin must be < {}".format(niters))
    if thin < 1:
        raise ValueError("thin must be >= 1")
    if thin >= niters-burnin:
        raise ValueError("thin must be < {}".format(niters-burnin))


    # lnprobabilities[iteration, walker]
    lnprobabilities = lnprobabilities[burnin:, :][::thin, :]
    # samples[iteration, walker, parameter]
    samples = samples[burnin:, :, :][::thin, : :]
    if adopt_inds is not None:
        samples = samples[:, :, adopt_inds]

    if flatten:
        lnprob_inds = np.where(lnprobabilities > lnprob_cutoff)
        samples = samples[lnprob_inds]
        lnprobabilities = lnprobabilities[lnprob_inds]
        # samples 2D (n, nparameters)
        # lnprobabilities 1D (n)
    else:
        lnprobabilities = _deepcopy(lnprobabilities)
        samples = _deepcopy(samples)

        samples[lnprobabilities < lnprob_cutoff] = np.nan
        lnprobabilities[lnprobabilities < lnprob_cutoff] = np.nan

        # lnprobabilities 2D (niters (after burnin, thin), nwalkers), replaced with nans where lnprob_cutoff applies
        # samples 3D (niters (after burning,thin), nwalkers, nparameters)

    return lnprobabilities, samples

def get_emcee_object_from_solution(b, solution, adopt_parameters=None):
    """
    Expose the `EnsembleSampler` object in `emcee` from the solution <phoebe.parameters.ParameterSet>.

    See also:
    * <phoebe.helpers.get_emcee_object>
    * <phoebe.parameters.solver.sampler.emcee>

    Arguments
    ------------
    * `b` (<phoebe.frontend.bundle.Bundle>): the Bundle
    * `solution` (string): solution label with `kind=='dynesty'`
    * `adopt_parameters` (list, optional, default=None): If not None, will
        override the value of `adopt_parameters` in the solution.

    Returns
    -----------
    * [emcee.EnsembleSampler](https://emcee.readthedocs.io/en/stable/user/sampler/#emcee.EnsembleSampler) object
    """
    solution_ps = b.get_solution(solution=solution, **_skip_filter_checks)
    if solution_ps.kind != 'emcee':
        raise ValueError("solution_ps must have kind 'emcee'")

    adopt_inds, adopt_uniqueids = b._get_adopt_inds_uniqueids(solution_ps, adopt_parameters=adopt_parameters)

    samples = solution_ps.get_value(qualifier='samples', **_skip_filter_checks) # shape: (niters, nwalkers, nparams)
    lnprobabilites = solution_ps.get_value(qualifier='lnprobabilities', **_skip_filter_checks) # shape: (niters, nwalkers)
    acceptance_fractions = solution_ps.get_value(qualifier='acceptance_fractions', **_skip_filter_checks) # shape: (nwalkers)

    return get_emcee_object(samples[:,:,adopt_inds], lnprobabilites, acceptance_fractions)

def get_emcee_object(samples, lnprobabilities, acceptance_fractions):
    """
    Expose the `EnsembleSampler` object in `emcee`.

    See also:
    * <phoebe.helpers.get_emcee_object_from_solution>

    Arguments
    ------------
    * `samples` (array): samples with shape (niters, nwalkers, nparams)
    * `lnprobabilities` (array): log-probablities with shape (niters, nwalkers)
    * `acceptance_fractions` (array): acceptance fractions with shape (nwalkers)

    Returns
    -----------
    * [emcee.EnsembleSampler](https://emcee.readthedocs.io/en/stable/user/sampler/#emcee.EnsembleSampler) object
    """
    if not _use_emcee:
        raise ImportError("emcee is not installed")

    backend = emcee.backends.Backend()
    backend.nwalkers = samples.shape[1]
    backend.ndim = samples.shape[2]
    backend.iteration = samples.shape[0]
    backend.accepted = acceptance_fractions
    backend.chain = samples
    backend.log_prob = lnprobabilities
    backend.initialized = True
    backend.random_state = None
    if not hasattr(backend, 'blobs'):
        # some versions of emcee seem to have a bug where it tries
        # to access backend.blobs but that does not exist.  Since
        # we don't use blobs, we'll get around that by faking it to
        # be None
        backend.blobs = None

    return backend


def get_dynesty_object_from_solution(b, solution, adopt_parameters=None):
    """
    Expose the `results` object in `dynesty` from the solution <phoebe.parameters.ParameterSet>.

    See also:
    * <phoebe.helpers.get_dynesty_object>
    * <phoebe.parameters.solver.sampler.dynesty>

    Arguments
    ------------
    * `b` (<phoebe.frontend.bundle.Bundle>): the Bundle
    * `solution` (string): solution label with `kind=='dynesty'`
    * `adopt_parameters` (list, optional, default=None): If not None, will
        override the value of `adopt_parameters` in the solution.

    Returns
    -----------
    * [dynesty.results.Results](https://dynesty.readthedocs.io/en/latest/api.html#module-dynesty.results) object
    """
    solution_ps = b.get_solution(solution=solution, **_skip_filter_checks)
    if solution_ps.kind != 'dynesty':
        raise ValueError("solution_ps must have kind 'dynesty'")

    adopt_inds, adopt_uniqueids = b._get_adopt_inds_uniqueids(solution_ps, adopt_parameters=adopt_parameters)

    return get_dynesty_object(adopt_inds=adopt_inds, **{p.qualifier: p.get_value() for p in solution_ps.to_list() if p.qualifier in ['nlive', 'niter', 'ncall', 'eff', 'samples', 'samples_id', 'samples_it', 'samples_u', 'logwt', 'logl', 'logvol', 'logz', 'logzerr', 'information', 'bound_iter', 'samples_bound', 'scale']})

def get_dynesty_object(nlive, niter, ncall, eff,
                       samples, samples_id, samples_it, samples_u,
                       logwt, logl, logvol, logz, logzerr,
                       information, bound_iter,
                       samples_bound, scale,
                       adopt_inds=None):
    """
    Expose the `results` object in `dynesty`.

    This can then be passed to any [dynesty helper function](https://dynesty.readthedocs.io/en/latest/api.html#module-dynesty.results)
    that takes `results`.

    See also:
    * <phoebe.helpers.get_dynesty_object_from_solution>

    Arguments
    ------------
    * positional arguments: all positional arguments are passed directly to dynesty.
    * `adopt_inds` (list, optional, default=None): If not None, only include
        the parameters with `adopt_inds` indices from `samples` and `samples_u`.

    """
    if not _use_dynesty:
        raise ImportError("dynesty is not installed")

    def _filter_by_adopt_inds(p, adopt_inds):
        if adopt_inds is not None and p.qualifier in ['samples', 'samples_u']:
            return p.value[:, adopt_inds]
        return p.value

    return _dynesty.results.Results(nlive=nlive, niter=niter, ncall=ncall, eff=eff,
                                    samples=samples, samples_id=samples_id,
                                    samples_it=samples_it, samples_u=samples_u,
                                    logwt=logwt, logl=logl, logvol=logvol, logz=logz, logzerr=logzerr,
                                    information=information, bound_iter=bound_iter,
                                    samples_bound=samples_bound, scale=scale)




from . import nbody as nbody
from . import keplerian as keplerian
from phoebe import u


# def dynamics_at_i(xs, ys, zs, vxs, vys, vzs, ethetas=None, elongans=None, eincls=None, i=0):
#     xi, yi, zi = [x[i].to(u.solRad) for x in xs], [y[i].to(u.solRad) for y in ys], [z[i].to(u.solRad) for z in zs]
#     vxi, vyi, vzi = [vx[i].to(u.solRad/u.d) for vx in vxs], [vy[i].to(u.solRad/u.d) for vy in vys], [vz[i].to(u.solRad/u.d) for vz in vzs]
#     if ethetas is not None:
#         ethetai, elongani, eincli = [etheta[i].to(u.rad) for etheta in ethetas], [elongan[i].to(u.rad) for elongan in elongans], [eincl[i].to(u.rad) for eincl in eincls]
#     else:
#         ethetai, elongani, eincli = None, None, None

#     return xi, yi, zi, vxi, vyi, vzi, ethetai, elongani, eincli

def at_i(array, i=0):
    return [a[i] for a in array]


def dynamics_at_i(xs, ys, zs, vxs, vys, vzs, ethetas=None, elongans=None, eincls=None, i=0):
    # xi, yi, zi = [x[i] for x in xs], [y[i] for y in ys], [z[i] for z in zs]
    xi, yi, zi = at_i(xs, i), at_i(ys, i), at_i(zs, i)
    # vxi, vyi, vzi = [vx[i] for vx in vxs], [vy[i] for vy in vys], [vz[i] for vz in vzs]
    vxi, vyi, vzi = at_i(vxs, i), at_i(vys, i), at_i(vzs, i)
    if ethetas is not None:
        # ethetai, elongani, eincli = [etheta[i] for etheta in ethetas], [elongan[i] for elongan in elongans], [eincl[i] for eincl in eincls]
        ethetai, elongani, eincli = at_i(ethetas, i), at_i(elongans, i), at_i(eincls, i)
    else:
        ethetai, elongani, eincli = None, None, None

    return xi, yi, zi, vxi, vyi, vzi, ethetai, elongani, eincli




from phoebe.parameters import *
from phoebe.parameters import constraint
from phoebe.atmospheres import passbands  # need to load pbtable (dictionary of available passbands)
from phoebe import u
from phoebe import conf

### NOTE: if creating new parameters, add to the _forbidden_labels list in parameters.py

_ld_func_choices = ['linear', 'logarithmic', 'quadratic', 'square_root', 'power']


passbands._init_passbands()  # TODO: move to module import
_ld_coeffs_source_choices = ['auto'] + list(set([atm for pb in passbands._pbtable.values() for atm in pb['atms_ld']]))

global _mesh_columns
global _pbdep_columns

_mesh_columns = []
# _mesh_columns += ['pot', 'rpole']
_mesh_columns += ['volume']

_mesh_columns += ['xs', 'ys', 'zs']
_mesh_columns += ['vxs', 'vys', 'vzs']
_mesh_columns += ['nxs', 'nys', 'nzs']

_mesh_columns += ['us', 'vs', 'ws']
_mesh_columns += ['vus', 'vvs', 'vws']
_mesh_columns += ['nus', 'nvs', 'nws']

# _mesh_columns += ['horizon_xs', 'horizon_ys', 'horizon_zs', 'horizon_analytic_xs', 'horizon_analytic_ys', 'horizon_analytic_zs']
_mesh_columns += ['areas'] #, 'tareas']
_mesh_columns += ['loggs', 'teffs']

_mesh_columns += ['rprojs', 'mus', 'visibilities', 'visible_centroids']
_mesh_columns += ['rs'] #, 'cosbetas']

lc_columns = []
lc_columns += ['intensities', 'normal_intensities', 'abs_intensities', 'abs_normal_intensities']
lc_columns += ['boost_factors', 'ldint']
lc_columns += ['pblum_ext', 'abs_pblum_ext', 'ptfarea']

# rv columns have all pb-dependent columns except those that require pblum scaling
rv_columns = [c for c in lc_columns[:] if c not in ['intensities', 'normal_intensities', 'pblum_ext']]
rv_columns += ['rvs']

lp_columns = rv_columns[:]
# lp_columns += ['dls']


_pbdep_columns = {'lc': lc_columns,
                  'rv': rv_columns,
                  'lp': lp_columns}

import logging
logger = logging.getLogger("DATASET")
logger.addHandler(logging.NullHandler())

def _empty_array(kwargs, qualifier):
    if qualifier in kwargs.keys():
        return kwargs.get(qualifier)
    elif 'empty_arrays_len' in kwargs.keys():
        return np.empty(kwargs.get('empty_arrays_len'))
    else:
        return []

def lc(syn=False, as_ps=True, is_lc=True, **kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for a light curve dataset.

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_dataset> as
    `b.add_dataset('lc')`.  In this case, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    Arguments
    ----------
    * `syn` (bool, optional, default=False): whether to create the parameters
        for the synthetic (model) instead of the observational (dataset).
    * `as_ps` (bool, optional, default=True): whether to return the parameters
        as a <phoebe.parameters.ParameterSet> instead of a list of
        <phoebe.parameters.Parameter> objects.
    * `is_lc` (bool, optional, default=True): whether to build all lc parameters
        or just passband-dependent parameters.  Generally this is only used
        internally by other dataset types that need passband-dependent parameters.
    * `times` (array/quantity, optional): observed times.  Only applicable
        if `is_lc` is True.
    * `fluxes` (array/quantity, optional): observed flux.  Only applicable
        if `is_lc` is True.
    * `sigmas` (array/quantity, optional): errors on flux measurements.  Only
        applicable if `syn` is False and `is_lc` is True.
    * `compute_times` (array/quantity, optional): times at which to compute
        the model.  Only applicable if `syn` is False and `is_lc` is True.
    * `compute_phases` (array/quantity, optional): phases at which to compute
        the model.  Only applicable if `syn` is False and `is_lc` is True.
    * `phases_period` (string, optional, default='period'): period to use
        when converting between `compute_phases` and `compute_times` as well as
        when applying `mask_phases`.  Only applicable if `syn` is False. and `is_lc` is True.
        Not applicable for single stars (in which case period is always used)
        or if `dperdt == 0`.
    * `phases_dpdt` (string, optional, default='dpdt'): dpdt to use when
        converting between compute_times and compute_phases as well as when
        applying mask_phases.  Not applicable for single stars or if `dpdt == 0`
    * `phases_t0` (string, optional, default='t0_supconj'): t0 to use
        when converting between `compute_phases` and `compute_times` as well as
        when applying `mask_phases`.  Only
        applicable if `syn` is False and `is_lc` is True.  Not applicable for
        single stars (in which case t0@system is always used).
    * `mask_enabled` (bool, optional, default=True): whether to apply the mask
        in mask_phases during plotting, calculate_residuals, calculate_chi2,
        calculate_lnlikelihood, and run_solver
    * `mask_phases` (list of tuples, optional, default=[]): List of phase-tuples.
        Any observations inside the range set by any of the tuples will be included.
    * `ld_mode` (string, optional, default='interp'): mode to use for handling
        limb-darkening.  Note that 'interp' is not available for all values
        of `atm` (availability can be checked by calling
        <phoebe.frontend.bundle.Bundle.run_checks> and will automatically be checked
        during <phoebe.frontend.bundle.Bundle.run_compute>).  Only applicable
        if `syn` is False.
    * `ld_func` (string, optional, default='logarithmic'): function/law to use for
        limb-darkening model. Not applicable if `ld_mode` is 'interp'.  Only
        applicable if `syn` is False.
    * `ld_coeffs_source` (string, optional, default='auto'): source for limb-darkening
        coefficients ('auto' to interpolate from the applicable table according
        to the 'atm' parameter, or the name of a specific atmosphere table).
        Only applicable if `ld_mode` is 'lookup'.  Only applicable if
        `syn` is False.
    * `ld_coeffs` (list, optional): limb-darkening coefficients.  Must be of
        the approriate length given the value of `ld_coeffs_source` which can
        be checked by calling <phoebe.frontend.bundle.Bundle.run_checks>
        and will automtically be checked during
        <phoebe.frontend.bundle.Bundle.run_compute>.  Only applicable
       if `ld_mode` is 'manual'.  Only applicable if `syn` is False.
    * `passband` (string, optional): passband.  Only applicable if `syn` is False.
    * `intens_weighting` (string, optional): whether passband intensities are
        weighted by energy or photons.  Only applicable if `syn` is False.
    * `pblum_mode` (string, optional, default='manual'): mode for scaling
        passband luminosities.  Only applicable if `syn` is False and `is_lc`
        is True.
    * `pblum_dataset` (string, optional):  Dataset to reference for coupling
        luminosities.  Only applicable if `pblum_mode` is 'dataset-coupled'.
    * `pblum_component` (string, optional): Component to provide `pblum`.
        Only applicable if `pblum_mode` is 'component-coupled'.
    * `pblum` (float/quantity or string, optional): passband luminosity (defined at t0).
        If `pblum_mode` is 'decoupled', then one entry per-star will be applicable.
        If `pblum_mode` is 'component-coupled', then only the entry for the primary
        component, according to <phoebe.parameters.HierarchyParameter> will be
        available.  To change the provided component, see `pblum_component`.
        Only applicable if `syn` is False and `is_lc` is True.
    * `l3_mode` (string, optional, default='flux'): mode for providing third
        light (`l3`).  Only applicable if `syn` is False and `is_lc` is True.
    * `l3` (float/quantity, optional): third light in flux units (only applicable
        if `l3_mode` is 'flux'). Only applicable if `syn` is False and `is_lc`
        is True.
    * `l3_frac` (float/quantity, optional): third light in fraction of total light.
        (only applicable if `l3_mode` is 'fraction').
        Only applicable if `syn` is False and `is_lc` is True.
    * `exptime` (float/quantity, optional): exposure time of the observations
        (`times` is defined at the mid-exposure).
        Only applicable if `syn` is False and `is_lc` is True.

    Returns
    --------
    * (<phoebe.parameters.ParameterSet> or list, list): ParameterSet (if `as_ps`)
        or list of all newly created
        <phoebe.parameters.Parameter> objects and a list of all necessary
        constraints.
    """

    params, constraints = [], []


    if is_lc:
        params += [FloatArrayParameter(qualifier='times', value=kwargs.get('times', []), required_shape=[None], readonly=syn, default_unit=u.d, description='Model (synthetic) times' if syn else 'Observed times')]
        params += [FloatArrayParameter(qualifier='fluxes', value=_empty_array(kwargs, 'fluxes'), required_shape=[None] if not syn else None, readonly=syn, default_unit=u.W/u.m**2, description='Model (synthetic) flux' if syn else 'Observed flux')]

    if not syn:
        # TODO: should we move all limb-darkening to compute options since
        # not all backends support interp (and lookup is atm-dependent)
        params += [ChoiceParameter(qualifier='ld_mode', copy_for={'kind': ['star'], 'component': '*'}, component='_default',
                                   value=kwargs.get('ld_mode', 'interp'), choices=['interp', 'lookup', 'manual'],
                                   description='Mode to use for limb-darkening')]
        params += [ChoiceParameter(visible_if='ld_mode:lookup|manual', qualifier='ld_func',
                                   copy_for={'kind': ['star'], 'component': '*'}, component='_default',
                                   value=kwargs.get('ld_func', 'logarithmic'), choices=_ld_func_choices,
                                   description='Limb darkening model')]
        params += [ChoiceParameter(visible_if='ld_mode:lookup', qualifier='ld_coeffs_source',
                                   copy_for={'kind': ['star'], 'component': '*'}, component='_default',
                                   value=kwargs.get('ld_coeffs_source', 'auto'), choices=_ld_coeffs_source_choices,
                                   advanced=True,
                                   description='Source for limb darkening coefficients (\'auto\' to interpolate from the applicable table according to the \'atm\' parameter, or the name of a specific atmosphere table)')]
        params += [FloatArrayParameter(visible_if='ld_mode:manual', qualifier='ld_coeffs',
                                       latexfmt=r'\mathrm{{ ldc }}_\mathrm{{ {dataset}, {component} }}',
                                       copy_for={'kind': ['star'], 'component': '*'}, component='_default',
                                       value=kwargs.get('ld_coeffs', [0.5, 0.5]), default_unit=u.dimensionless_unscaled,
                                       required_shape=[None],
                                       description='Limb darkening coefficients')]

        passbands._init_passbands()  # NOTE: this only actually does something on the first call
        params += [ChoiceParameter(qualifier='passband', value=kwargs.get('passband', 'Johnson:V'), choices=passbands.list_passbands(), description='Passband')]
        params += [ChoiceParameter(qualifier='intens_weighting', value=kwargs.get('intens_weighting', 'energy'), choices=['energy', 'photon'], advanced=True, description='Whether passband intensities are weighted by energy or photons')]



    if is_lc and not syn:
        params += [FloatArrayParameter(qualifier='compute_times', value=kwargs.get('compute_times', []), required_shape=[None], default_unit=u.d, description='Times to use during run_compute.  If empty, will use times parameter')]
        params += [FloatArrayParameter(qualifier='compute_phases', component=kwargs.get('component_top', None), required_shape=[None], value=kwargs.get('compute_phases', []), default_unit=u.dimensionless_unscaled, description='Phases associated with compute_times.')]
        params += [ChoiceParameter(qualifier='phases_period', visible_if='[dataset][context][kind]dperdt:!0.0,hierarchy.is_meshable:False', component=kwargs.get('component_top', None), value=kwargs.get('phases_period', 'period'), choices=['period', 'period_anom'], advanced=True, description='period to use when converting between compute_times and compute_phases as well as when applying mask_phases')]
        params += [ChoiceParameter(qualifier='phases_dpdt', visible_if='[dataset][context][kind]dpdt:!0.0,hierarchy.is_meshable:False', component=kwargs.get('component_top', None), value=kwargs.get('phases_dpdt', 'dpdt'), choices=['dpdt', 'none'], advanced=True, description='dpdt to use when converting between compute_times and compute_phases as well as when applying mask_phases')]
        params += [ChoiceParameter(qualifier='phases_t0', visible_if='hierarchy.is_meshable:False', component=kwargs.get('component_top', None), value=kwargs.get('phases_t0', 't0_supconj'), choices=['t0_supconj', 't0_perpass', 't0_ref'], advanced=True, description='t0 to use when converting between compute_times and compute_phases as well as when applying mask_phases')]
        constraints += [(constraint.compute_phases, kwargs.get('component_top', None), kwargs.get('dataset', None))]

        params += [BoolParameter(qualifier='mask_enabled', value=kwargs.get('mask_enabled', True), description='Whether to apply the mask in mask_phases during plotting, calculate_residuals, calculate_chi2, calculate_lnlikelihood, and run_solver')]
        params += [FloatArrayParameter(visible_if='[component]mask_enabled:True', qualifier='mask_phases', component=kwargs.get('component_top', None), value=kwargs.get('mask_phases', []), default_unit=u.dimensionless_unscaled, required_shape=[None, 2], description='List of phase-tuples.  Any observations inside the range set by any of the tuples will be included.')]

        params += [ChoiceParameter(qualifier='solver_times', value=kwargs.get('solver_times', 'auto'), choices=['auto', 'compute_times', 'times'], description='times to use within run_solver.  All options will properly account for masking from mask_times.  To see how this is parsed, see b.parse_solver_times.  auto: use compute_times if provided and shorter than times, otherwise use times.  compute_times: use compute_times if provided.  times: use times array.')]

        params += [FloatArrayParameter(qualifier='sigmas', value=_empty_array(kwargs, 'sigmas'), required_shape=[None], default_unit=u.W/u.m**2, description='Observed uncertainty on flux')]
        params += [FloatParameter(qualifier='sigmas_lnf', latexfmt=r'\sigma_\mathrm{{ lnf, {dataset} }}', visible_if='sigmas:<notempty>', value=kwargs.get('sigmas_lnf', -np.inf), default_unit=u.dimensionless_unscaled, limits=(None, None), description='Natural log of the fractional amount to sigmas are underestimate (when calculating chi2/lnlikelihood)')]

        params += [ChoiceParameter(qualifier='pblum_mode', value=kwargs.get('pblum_mode', 'component-coupled'),
                                   choices=['decoupled', 'component-coupled', 'dataset-coupled', 'dataset-scaled', 'absolute'],
                                   description='Mode for scaling passband luminosities')]

        # pblum_mode = 'component-coupled' or 'decoupled'
        params += [ChoiceParameter(visible_if='pblum_mode:component-coupled', qualifier='pblum_component', value=kwargs.get('pblum_component', ''), choices=kwargs.get('starrefs', ['']), advanced=True, description='Which component\'s pblum will be provided')]
        params += [FloatParameter(qualifier='pblum', latexfmt=r'L_\mathrm{{ pb, {dataset} }}', visible_if='[component]pblum_mode:decoupled||[component]pblum_mode:component-coupled,[component]pblum_component:<component>', copy_for={'kind': ['star'], 'component': '*'}, component='_default', value=kwargs.get('pblum', 4*np.pi), default_unit=u.W, description='Passband luminosity (defined at t0)')]

        # pblum_mode = 'dataset-coupled'
        params += [ChoiceParameter(visible_if='pblum_mode:dataset-coupled', qualifier='pblum_dataset', value=kwargs.get('pblum_dataset', ''), choices=['']+kwargs.get('lcrefs', []), description='Dataset with which to couple luminosities based on color')]

        # pblum_mode = 'pbflux'
        params += [FloatParameter(visible_if='false', qualifier='pbflux', value=kwargs.get('pbflux', 1.0), default_unit=u.W/u.m**2, description='Total inrinsic (excluding features and irradiation) passband flux (at t0, including l3 if pblum_mode=\'flux\')')]

        params += [ChoiceParameter(qualifier='l3_mode', value=kwargs.get('l3_mode', 'flux'), choices=['flux', 'fraction'], description='Whether third light is given in units of flux or as a fraction of total light')]
        params += [FloatParameter(visible_if='l3_mode:flux', qualifier='l3', latexfmt=r'l_\mathrm{{ 3, {dataset} }}', value=kwargs.get('l3', 0.), limits=[0, None], default_unit=u.W/u.m**2, description='Third light in flux units')]
        params += [FloatParameter(visible_if='l3_mode:fraction', qualifier='l3_frac', latexfmt=r'l_\mathrm{{ 3, frac, {dataset} }}', value=kwargs.get('l3_frac', 0.), limits=[0, 1], default_unit=u.dimensionless_unscaled, description='Third light as a fraction of total flux (both system and third light)')]

        params += [FloatParameter(qualifier='exptime', value=kwargs.get('exptime', 0.0), default_unit=u.s, description='Exposure time (time is defined as mid-exposure)')]


    return ParameterSet(params) if as_ps else params, constraints

def rv(syn=False, as_ps=True, **kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for a radial velocity dataset.

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_dataset> as
    `b.add_dataset('rv')`.  In this case, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    Arguments
    ----------
    * `syn` (bool, optional, default=False): whether to create the parameters
        for the synthetic (model) instead of the observational (dataset).
    * `as_ps` (bool, optional, default=True): whether to return the parameters
        as a <phoebe.parameters.ParameterSet> instead of a list of
        <phoebe.parameters.Parameter> objects.
    * `times` (array/quantity, optional): observed times.
    * `rvs` (array/quantity, optional): observed radial velocities.
    * `sigmas` (array/quantity, optional): errors on radial velocity measurements.
        Only applicable if `syn` is False.
    * `compute_times` (array/quantity, optional): times at which to compute
        the model.  Only applicable if `syn` is False.
    * `compute_phases` (array/quantity, optional): phases at which to compute
        the model.  Only applicable if `syn` is False.
    * `phases_period` (string, optional, default='period'): period to use
        when converting between `compute_phases` and `compute_times` as well as
        when applying `mask_phases`.  Only applicable if `syn` is False.  Not applicable for
        single stars (in which case period is always used) or if `dperdt == 0.0`.
    * `phases_dpdt` (string, optional, default='dpdt'): dpdt to use when
        converting between compute_times and compute_phases as well as when
        applying mask_phases.  Not applicable for single stars or if `dpdt == 0`
    * `phases_t0` (string, optional, default='t0_supconj'): t0 to use
        when converting between `compute_phases` and `compute_times` as well as
        when applying `mask_phases`.  Only applicable if `syn` is False.  Not applicable for
        single stars (in which case t0@system is always used).
    * `mask_enabled` (bool, optional, default=True): whether to apply the mask
        in mask_phases during plotting, calculate_residuals, calculate_chi2,
        calculate_lnlikelihood, and run_solver
    * `mask_phases` (list of tuples, optional, default=[]): List of phase-tuples.
        Any observations inside the range set by any of the tuples will be included.
    * `ld_mode` (string, optional, default='interp'): mode to use for handling
        limb-darkening.  Note that 'interp' is not available for all values
        of `atm` (availability can be checked by calling
        <phoebe.frontend.bundle.Bundle.run_checks> and will automatically be checked
        during <phoebe.frontend.bundle.Bundle.run_compute>).  Only applicable
        if `syn` is False.
    * `ld_func` (string, optional, default='linear'): function/law to use for
        limb-darkening model. Not applicable if `ld_mode` is 'interp'.  Only
        applicable if `syn` is False.
    * `ld_coeffs_source` (string, optional, default='auto'): source for limb-darkening
        coefficients ('auto' to interpolate from the applicable table according
        to the 'atm' parameter, or the name of a specific atmosphere table).
        Only applicable if `ld_mode` is 'lookup'.  Only applicable if
        `syn` is False.
    * `ld_coeffs` (list, optional): limb-darkening coefficients.  Must be of
        the approriate length given the value of `ld_coeffs_source` which can
        be checked by calling <phoebe.frontend.bundle.Bundle.run_checks>
        and will automtically be checked during
        <phoebe.frontend.bundle.Bundle.run_compute>.  Only applicable
       if `ld_mode` is 'manual'.  Only applicable if `syn` is False.
    * `passband` (string, optional): passband.  Only applicable if `syn` is False.
    * `intens_weighting` (string, optional): whether passband intensities are
        weighted by energy or photons.  Only applicable if `syn` is False.

    Returns
    --------
    * (<phoebe.parameters.ParameterSet> or list, list): ParameterSet (if `as_ps`)
        or list of all newly created
        <phoebe.parameters.Parameter> objects and a list of all necessary
        constraints.
    """

    params, constraints = [], []

    params += [FloatArrayParameter(qualifier='times', copy_for={'kind': ['star'], 'component': '*'}, component='_default', value=kwargs.get('times', []), required_shape=[None], readonly=syn, default_unit=u.d, description='Model (synthetic) times' if syn else 'Observed times')]
    params += [FloatArrayParameter(qualifier='rvs', visible_if='times:<notempty>', copy_for={'kind': ['star'], 'component': '*'}, component='_default', value=_empty_array(kwargs, 'rvs'), required_shape=[None] if not syn else None, readonly=syn, default_unit=u.km/u.s, description='Model (synthetic) radial velocities' if syn else 'Observed radial velocity')]

    if not syn:
        params += [FloatArrayParameter(qualifier='sigmas', visible_if='times:<notempty>', copy_for={'kind': ['star'], 'component': '*'}, component='_default', value=_empty_array(kwargs, 'sigmas'), required_shape=None, default_unit=u.km/u.s, description='Observed uncertainty on rv')]
        params += [FloatParameter(qualifier='sigmas_lnf', latexfmt=r'\sigma_\mathrm{{ lnf, {dataset} }}', visible_if='sigmas:<notempty>', copy_for={'kind': ['star'], 'component': '*'}, component='_default', value=kwargs.get('sigmas_lnf', -np.inf), default_unit=u.dimensionless_unscaled, limits=(None,None), description='Natural log of the fractional amount to sigmas are underestimate (when calculating chi2/lnlikelihood)')]

        params += [FloatParameter(qualifier='rv_offset', copy_for={'kind': ['star'], 'component': '*'}, component='_default', value=kwargs.get('rv_offset', 0.0), default_unit=u.km/u.s, description='Per-component offset to add to synthetic RVs (i.e. for hot stars)')]

        params += [FloatArrayParameter(qualifier='compute_times', value=kwargs.get('compute_times', []), required_shape=[None], default_unit=u.d, description='Times to use during run_compute.  If empty, will use times parameter')]
        params += [FloatArrayParameter(qualifier='compute_phases', component=kwargs.get('component_top', None), value=kwargs.get('compute_phases', []), required_shape=[None], default_unit=u.dimensionless_unscaled, description='Phases associated with compute_times.')]
        params += [ChoiceParameter(qualifier='phases_period', visible_if='[dataset][context][kind]dperdt:!0.0,hierarchy.is_meshable:False', component=kwargs.get('component_top', None), value=kwargs.get('phases_period', 'period'), choices=['period', 'period_anom'], advanced=True, description='period to use when converting between compute_times and compute_phases as well as when applying mask_phases')]
        params += [ChoiceParameter(qualifier='phases_dpdt', visible_if='[dataset][context][kind]dpdt:!0.0,hierarchy.is_meshable:False', component=kwargs.get('component_top', None), value=kwargs.get('phases_dpdt', 'dpdt'), choices=['dpdt', 'none'], advanced=True, description='dpdt to use when converting between compute_times and compute_phases as well as when applying mask_phases')]
        params += [ChoiceParameter(qualifier='phases_t0', visible_if='hierarchy.is_meshable:False', component=kwargs.get('component_top', None), value=kwargs.get('phases_t0', 't0_supconj'), choices=['t0_supconj', 't0_perpass', 't0_ref'], advanced=True, description='t0 to use when converting between compute_times and compute_phases as well as when applying mask_phases')]
        constraints += [(constraint.compute_phases, kwargs.get('component_top', None), kwargs.get('dataset', None))]

        params += [BoolParameter(qualifier='mask_enabled', value=kwargs.get('mask_enabled', True), description='Whether to apply the mask in mask_phases during plotting, calculate_residuals, calculate_chi2, calculate_lnlikelihood, and run_solver')]
        params += [FloatArrayParameter(visible_if='[component]mask_enabled:True', qualifier='mask_phases', component=kwargs.get('component_top', None), value=kwargs.get('mask_phases', []), default_unit=u.dimensionless_unscaled, required_shape=[None, 2], description='List of phase-tuples.  Any observations inside the range set by any of the tuples will be included.')]

        params += [ChoiceParameter(qualifier='solver_times', value=kwargs.get('solver_times', 'auto'), choices=['auto', 'compute_times', 'times'], description='times to use within run_solver.  auto: use compute_times if provided and shorter than times, otherwise use times.  compute_times: use compute_times if provided.  times: use times array.')]


    lc_params, lc_constraints = lc(syn=syn, as_ps=False, is_lc=False, **kwargs)
    params += lc_params
    constraints += lc_constraints

    return ParameterSet(params) if as_ps else params, constraints

def lp(syn=False, as_ps=True, **kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for a line profile dataset.

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_dataset> as
    `b.add_dataset('lp', times=[...])`.  In this case, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    Note that `times` **must** be passed during creation and cannot be changed
    after the fact as this function creates copies of the `flux_densities`
    and `sigmas` parameters per-time instead of creating a `times` parameter.

    Arguments
    ----------
    * `syn` (bool, optional, default=False): whether to create the parameters
        for the synthetic (model) instead of the observational (dataset).
    * `as_ps` (bool, optional, default=True): whether to return the parameters
        as a <phoebe.parameters.ParameterSet> instead of a list of
        <phoebe.parameters.Parameter> objects.
    * `times` (array/quantity): times at which the dataset should be defined.
        **IMPORTANT**: times is not a parameter and must be passed during creation,
        see note above.  If `syn` is True, a `times` parameter will be created,
        but all other parameters will be tagged with individual times.
    * `wavelengths` (array/quantity, optional): observed wavelengths.
    * `flux_densities` (array/quantity, optional): observed flux densities.
        A copy of this parameter will exist per-time (as passed to the `times`
        argument at creation, see above) and will be tagged with that time.
    * `sigmas` (array/quantity, optional): errors on flux densities measurements.
        Only applicable if `syn` is False.  A copy of this parameter will exist
        per-time (as passed to the `times` argument at creation, see above) and
        will be tagged with that time.
    * `compute_times` (array/quantity, optional): times at which to compute
        the model.  If provided, this will override the tagged times as defined
        by `times` (note that interpolating between the model computed at
        `compute_times` and the dataset defined at `times` is not currently
        supported).  Only applicable if `syn` is False.
    * `compute_phases` (array/quantity, optional): phases at which to compute
        the model.  Only applicable if `syn` is False.
    * `phases_period` (string, optional, default='period'): period to use
        when converting between `compute_phases` and `compute_times` as well as
        when applying `mask_phases`.  Only applicable if `syn` is False.  Not applicable for
        single stars (in which case period is always used) or if `dperdt == 0`.
    * `phases_dpdt` (string, optional, default='dpdt'): dpdt to use when
        converting between compute_times and compute_phases as well as when
        applying mask_phases.  Not applicable for single stars or if `dpdt == 0`
    * `phases_t0` (string, optional, default='t0_supconj'): t0 to use
        when converting between `compute_phases` and `compute_times`.  Only
        applicable if `syn` is False.  Not applicable for
        single stars (in which case t0@system is always used).
    * `profile_func` (string, optional, default='gaussian'): function to use
        for the rest line profile.
    * `profile_rest` (float, optional, default=550): rest central wavelength
        for the line profile.
    * `profile_sv` (float, optional, default=1e-4): subsidiary value of the
        profile.
    * `ld_mode` (string, optional, default='interp'): mode to use for handling
        limb-darkening.  Note that 'interp' is not available for all values
        of `atm` (availability can be checked by calling
        <phoebe.frontend.bundle.Bundle.run_checks> and will automatically be checked
        during <phoebe.frontend.bundle.Bundle.run_compute>).  Only applicable
        if `syn` is False.
    * `ld_func` (string, optional, default='linear'): function/law to use for
        limb-darkening model. Not applicable if `ld_mode` is 'interp'.  Only
        applicable if `syn` is False.
    * `ld_coeffs_source` (string, optional, default='auto'): source for limb-darkening
        coefficients ('auto' to interpolate from the applicable table according
        to the 'atm' parameter, or the name of a specific atmosphere table).
        Only applicable if `ld_mode` is 'lookup'.  Only applicable if
        `syn` is False.
    * `ld_coeffs` (list, optional): limb-darkening coefficients.  Must be of
        the approriate length given the value of `ld_coeffs_source` which can
        be checked by calling <phoebe.frontend.bundle.Bundle.run_checks>
        and will automtically be checked during
        <phoebe.frontend.bundle.Bundle.run_compute>.  Only applicable
       if `ld_mode` is 'manual'.  Only applicable if `syn` is False.
    * `passband` (string, optional): passband.  Only applicable if `syn` is False.
    * `intens_weighting` (string, optional): whether passband intensities are
        weighted by energy or photons.  Only applicable if `syn` is False.

    Returns
    --------
    * (<phoebe.parameters.ParameterSet> or list, list): ParameterSet (if `as_ps`)
        or list of all newly created
        <phoebe.parameters.Parameter> objects and a list of all necessary
        constraints.
    """

    params, constraints = [], []

    times = kwargs.get('times', [])

    # if syn:
        # expose the computed times as we do for a mesh, even though the actual
        # parameters will be **tagged** with times
        # TODO: it would be nice if this wasn't copied per-component in the model... but it is also somewhat useful
        # NOTE: enabling this requires some changes to plotting logic.  We have it for meshes so you can plot pot vs time, for example, but lp doesn't have any syn FloatParameters
        # params += [FloatArrayParameter(qualifier='times', value=kwargs.get('times', []), default_unit=u.d, description='{} times'.format('Synthetic' if syn else 'Observed'))]


    # wavelengths is time-independent
    params += [FloatArrayParameter(qualifier='wavelengths', copy_for={'kind': ['star', 'orbit'], 'component': '*'}, component='_default', value=_empty_array(kwargs, 'wavelengths'), required_shape=[None], readonly=syn, default_unit=u.nm, description='Wavelengths of the model (synthetic)' if syn else 'Wavelengths of the observations')]

    for time in times:
        # but do allow per-component flux_densities and sigmas
        params += [FloatArrayParameter(qualifier='flux_densities', visible_if='[time]wavelengths:<notempty>', copy_for={'kind': ['star', 'orbit'], 'component': '*'},
                                       component='_default', time=time, value=_empty_array(kwargs, 'flux_densities'),
                                       required_shape=[None], readonly=syn, default_unit=u.W/(u.m**2*u.nm),
                                       description='Flux density per wavelength (must be same length as wavelengths or empty)')]

        if not syn:
            params += [FloatArrayParameter(qualifier='sigmas', visible_if='[time]wavelengths:<notempty>', copy_for={'kind': ['star', 'orbit'], 'component': '*'}, component='_default', time=time, value=_empty_array(kwargs, 'sigmas'), required_shape=[None], default_unit=u.W/(u.m**2*u.nm), description='Observed uncertainty on flux_densities')]
            params += [FloatParameter(qualifier='sigmas_lnf',  latexfmt=r'\sigma_\mathrm{{ lnf, {dataset} }}', visible_if='[time]sigmas:<notempty>', copy_for={'kind': ['star', 'orbit'], 'component': '*'}, component='_default', time=time, value=kwargs.get('sigmas_lnf', -np.inf), default_unit=u.dimensionless_unscaled, limits=(None, None), description='Natural log of the fractional amount to sigmas are underestimate (when calculating chi2/lnlikelihood)')]

    if not syn:
        params += [FloatArrayParameter(qualifier='compute_times', value=kwargs.get('compute_times', []), required_shape=[None], default_unit=u.d, description='Times to use during run_compute.  If empty, will use times of individual entries.  Note that interpolation is not currently supported for lp datasets.')]
        params += [FloatArrayParameter(qualifier='compute_phases', component=kwargs.get('component_top', None), value=kwargs.get('compute_phases', []), required_shape=[None], default_unit=u.dimensionless_unscaled, description='Phases associated with compute_times.')]
        params += [ChoiceParameter(qualifier='phases_period', visible_if='[dataset][context][kind]dperdt:!0.0,hierarchy.is_meshable:False', component=kwargs.get('component_top', None), value=kwargs.get('phases_period', 'period'), choices=['period', 'period_anom'], advanced=True, description='period to use when converting between compute_times and compute_phases as well as when applying mask_phases')]
        params += [ChoiceParameter(qualifier='phases_dpdt', visible_if='[dataset][context][kind]dpdt:!0.0,hierarchy.is_meshable:False', component=kwargs.get('component_top', None), value=kwargs.get('phases_dpdt', 'dpdt'), choices=['dpdt', 'none'], advanced=True, description='dpdt to use when converting between compute_times and compute_phases as well as when applying mask_phases')]
        params += [ChoiceParameter(qualifier='phases_t0', visible_if='hierarchy.is_meshable:False', component=kwargs.get('component_top', None), value=kwargs.get('phases_t0', 't0_supconj'), choices=['t0_supconj', 't0_perpass', 't0_ref'], advanced=True, description='t0 to use when converting between compute_times and compute_phases as well as when applying mask_phases')]
        constraints += [(constraint.compute_phases, kwargs.get('component_top', None), kwargs.get('dataset', None))]

        # params += [BoolParameter(qualifier='mask_enabled', value=kwargs.get('mask_enabled', True), description='Whether to apply the mask in mask_phases during plotting, calculate_residuals, calculate_chi2, calculate_lnlikelihood, and run_solver')]
        # params += [FloatArrayParameter(visible_if='[component]mask_enabled:True', qualifier='mask_phases', component=kwargs.get('component_top', None), value=kwargs.get('mask_phases', []), default_unit=u.dimensionless_unscaled, required_shape=[None, 2], description='List of phase-tuples.  Any observations inside the range set by any of the tuples will be included.')]

        params += [ChoiceParameter(qualifier='solver_times', value=kwargs.get('solver_times', 'auto'), choices=['auto', 'compute_times', 'times'], description='times to use within run_solver.  auto: use compute_times if provided and shorter than times, otherwise use times.  compute_times: use compute_times if provided.  times: use times array.')]

        params += [ChoiceParameter(qualifier='profile_func', value=kwargs.get('profile_func', 'gaussian'), choices=['gaussian', 'lorentzian'], description='Function to use for the rest line profile')]
        params += [FloatParameter(qualifier='profile_rest', value=kwargs.get('profile_rest', 550), default_unit=u.nm, limits=(0, None), description='Rest central wavelength of the profile')]
        params += [FloatParameter(qualifier='profile_sv', value=kwargs.get('profile_sv', 1e-4), default_unit=u.dimensionless_unscaled, limits=(0, None), description='Subsidiary value of the profile')]

    lc_params, lc_constraints = lc(syn=syn, as_ps=False, is_lc=False, **kwargs)
    params += lc_params
    constraints += lc_constraints

    return ParameterSet(params) if as_ps else params, constraints


def orb(syn=False, as_ps=True, **kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for an orbit dataset.

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_dataset> as
    `b.add_dataset('orb')`.  In this case, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    Arguments
    ----------
    * `syn` (bool, optional, default=False): whether to create the parameters
        for the synthetic (model) instead of the observational (dataset).
    * `as_ps` (bool, optional, default=True): whether to return the parameters
        as a <phoebe.parameters.ParameterSet> instead of a list of
        <phoebe.parameters.Parameter> objects.
    * `times` (array/quantity, optional): observed times.
    * `compute_times` (array/quantity, optional): times at which to compute
        the model.  Only applicable if `syn` is False.
    * `compute_phases` (array/quantity, optional): phases at which to compute
        the model.  Only applicable if `syn` is False.
    * `phases_period` (string, optional, default='period'): period to use
        when converting between `compute_phases` and `compute_times` as well as
        when applying `mask_phases`.  Only applicable if `syn` is False.  Not applicable for
        single stars (in which case period is always used) or if `dperdt == 0.0`.
    * `phases_dpdt` (string, optional, default='dpdt'): dpdt to use when
        converting between compute_times and compute_phases as well as when
        applying mask_phases.  Not applicable for single stars or if `dpdt == 0`
    * `phases_t0` (string, optional, default='t0_supconj'): t0 to use
        when converting between `compute_phases` and `compute_times`.  Only
        applicable if `syn` is False.  Not applicable for
        single stars (in which case t0@system is always used).

    Returns
    --------
    * (<phoebe.parameters.ParameterSet> or list, list): ParameterSet (if `as_ps`)
        or list of all newly created
        <phoebe.parameters.Parameter> objects and a list of all necessary
        constraints.
    """

    params, constraints = [], []

    if syn:
        params += [FloatArrayParameter(qualifier='times', copy_for={'kind': ['star'], 'component': '*'}, component='_default', value=kwargs.get('times', []), readonly=syn, default_unit=u.d, description='Model (synthetic) times' if syn else 'Observed times')]

    if syn:
        params += [FloatArrayParameter(qualifier='us', value=_empty_array(kwargs, 'us'), readonly=syn, default_unit=u.solRad, description='U position')]
        params += [FloatArrayParameter(qualifier='vs', value=_empty_array(kwargs, 'vs'), readonly=syn, default_unit=u.solRad, description='V position')]
        params += [FloatArrayParameter(qualifier='ws', value=_empty_array(kwargs, 'ws'), readonly=syn, default_unit=u.solRad, description='W position')]
        params += [FloatArrayParameter(qualifier='vus', value=_empty_array(kwargs, 'vus'), readonly=syn, default_unit=u.km/u.s, description='U velocity')]
        params += [FloatArrayParameter(qualifier='vvs', value=_empty_array(kwargs, 'vvs'), readonly=syn, default_unit=u.km/u.s, description='V velocity')]
        params += [FloatArrayParameter(qualifier='vws', value=_empty_array(kwargs, 'vws'), readonly=syn, default_unit=u.km/u.s, description='W velocity')]

    if not syn:
        params += [FloatArrayParameter(qualifier='compute_times', value=kwargs.get('compute_times', []), required_shape=[None], default_unit=u.d, description='Times to use during run_compute.  If empty, will use times parameter')]
        params += [FloatArrayParameter(qualifier='compute_phases', component=kwargs.get('component_top', None), value=kwargs.get('compute_phases', []), required_shape=[None], default_unit=u.dimensionless_unscaled, description='Phases associated with compute_times.')]
        params += [ChoiceParameter(qualifier='phases_period', visible_if='[dataset][context][kind]dperdt:!0.0,hierarchy.is_meshable:False', component=kwargs.get('component_top', None), value=kwargs.get('phases_period', 'period'), choices=['period', 'period_anom'], advanced=True, description='period to use when converting between compute_times and compute_phases as well as when applying mask_phases')]
        params += [ChoiceParameter(qualifier='phases_dpdt', visible_if='[dataset][context][kind]dpdt:!0.0,hierarchy.is_meshable:False', component=kwargs.get('component_top', None), value=kwargs.get('phases_dpdt', 'dpdt'), choices=['dpdt', 'none'], advanced=True, description='dpdt to use when converting between compute_times and compute_phases as well as when applying mask_phases')]
        params += [ChoiceParameter(qualifier='phases_t0', visible_if='hierarchy.is_meshable:False', component=kwargs.get('component_top', None), value=kwargs.get('phases_t0', 't0_supconj'), choices=['t0_supconj', 't0_perpass', 't0_ref'], advanced=True, description='t0 to use when converting between compute_times and compute_phases as well as when applying mask_phases')]
        constraints += [(constraint.compute_phases, kwargs.get('component_top', None), kwargs.get('dataset', None))]

    return ParameterSet(params) if as_ps else params, constraints

def mesh(syn=False, as_ps=True, **kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for a mesh dataset.

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_dataset> as
    `b.add_dataset('mesh')`.  In this case, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    Arguments
    ----------
    * `syn` (bool, optional, default=False): whether to create the parameters
        for the synthetic (model) instead of the observational (dataset).
    * `as_ps` (bool, optional, default=True): whether to return the parameters
        as a <phoebe.parameters.ParameterSet> instead of a list of
        <phoebe.parameters.Parameter> objects.
    * `times` (array/quantity, optional): observed times.  Only applicable
        if `syn` is True.  When `syn` is False: if provided, but `compute_times`
        is not provided, this will write to `compute_times` with a warning
        in the logger.
    * `compute_times` (array/quantity, optional): times at which to compute
        the model.  Only applicable if `syn` is False.
    * `compute_phases` (array/quantity, optional): phases at which to compute
        the model.  Only applicable if `syn` is False.
    * `phases_period` (string, optional, default='period'): period to use
        when converting between `compute_phases` and `compute_times` as well as
        when applying `mask_phases`.  Only applicable if `syn` is False.  Not applicable for
        single stars (in which case period is always used) or if `dperdt == 0.0`.
    * `phases_dpdt` (string, optional, default='dpdt'): dpdt to use when
        converting between compute_times and compute_phases as well as when
        applying mask_phases.  Not applicable for single stars or if `dpdt == 0`
    * `phases_t0` (string, optional, default='t0_supconj'): t0 to use
        when converting between `compute_phases` and `compute_times`.  Only
        applicable if `syn` is False.  Not applicable for
        single stars (in which case t0@system is always used).
    * `include_times` (string, optional): append to `compute_times` from the
        following datasets/time standards.  If referring to other datasets,
        this will copy the computed times of that dataset (whether that be
        from the `times` or `compute_times` of the respective dataset).
        Only applicable if `syn` is False.
    * `coordinates` (list, optional, default=['xyz', 'uvw']): coordinates to
        expose the mesh.  uvw (plane of sky) and/or xyz (roche).
    * `columns` (list, optional, default=[]): columns to expose within the mesh.
        Only applicable if `syn` is False.
    * `**kwargs`: if `syn` is True, additional kwargs will be applied to the
        exposed columns according to the passed lists for `mesh_columns`
        and `mesh_datasets`.

    Returns
    --------
    * (<phoebe.parameters.ParameterSet> or list, list): ParameterSet (if `as_ps`)
        or list of all newly created
        <phoebe.parameters.Parameter> objects and a list of all necessary
        constraints.
    """

    params, constraints = [], []

    times = kwargs.get('times', [])

    if syn:
        # TODO: it would be nice if this wasn't copied per-component in the model... but it is also somewhat useful
        params += [FloatArrayParameter(qualifier='times', value=kwargs.get('times', []), readonly=syn, default_unit=u.d, description='Model (synthetic) times' if syn else 'Observed times')]

    if not syn:
        if 'times' in kwargs.keys():
            if 'compute_times' in kwargs.keys():
                raise KeyError("mesh dataset cannot accept both 'times' and 'compute_times'.")
            else:
                logger.warning("mesh datasets do not have a 'times' parameter.  Applying value sent to 'times' to 'compute_times'")
                compute_times = kwargs.get('times', [])
        else:
            compute_times = kwargs.get('compute_times', [])


        params += [FloatArrayParameter(qualifier='compute_times', value=compute_times, required_shape=[None], default_unit=u.d, description='Times to use during run_compute.')]
        params += [FloatArrayParameter(qualifier='compute_phases', component=kwargs.get('component_top', None), value=kwargs.get('compute_phases', []), required_shape=[None], default_unit=u.dimensionless_unscaled, description='Phases associated with compute_times.')]
        params += [ChoiceParameter(qualifier='phases_period', visible_if='[dataset][context][kind]dperdt:!0.0,hierarchy.is_meshable:False', component=kwargs.get('component_top', None), value=kwargs.get('phases_period', 'period'), choices=['period', 'period_anom'], advanced=True, description='period to use when converting between compute_times and compute_phases as well as when applying mask_phases')]
        params += [ChoiceParameter(qualifier='phases_dpdt', visible_if='[dataset][context][kind]dpdt:!0.0,hierarchy.is_meshable:False', component=kwargs.get('component_top', None), value=kwargs.get('phases_dpdt', 'dpdt'), choices=['dpdt', 'none'], advanced=True, description='dpdt to use when converting between compute_times and compute_phases as well as when applying mask_phases')]
        params += [ChoiceParameter(qualifier='phases_t0', visible_if='hierarchy.is_meshable:False', component=kwargs.get('component_top', None), value=kwargs.get('phases_t0', 't0_supconj'), choices=['t0_supconj', 't0_perpass', 't0_ref'], advanced=True, description='t0 to use when converting between compute_times and compute_phases as well as when applying mask_phases')]
        constraints += [(constraint.compute_phases, kwargs.get('component_top', None), kwargs.get('dataset', None))]

        params += [SelectParameter(qualifier='include_times', value=kwargs.get('include_times', []), required_shape=[None], advanced=False, description='append to compute_times from the following datasets/time standards', choices=['t0@system'])]
        params += [SelectParameter(qualifier='coordinates', value=kwargs.get('coordinates', ['xyz', 'uvw']), choices=['xyz', 'uvw'], advanced=True, description='coordinates to expose the mesh.  uvw (plane of sky) and/or xyz (roche)')]
        params += [SelectParameter(qualifier='columns', value=kwargs.get('columns', []), description='columns to expose within the mesh', choices=_mesh_columns)]

    # the following will all be arrays (value per triangle) per time
    if syn:
        columns = kwargs.get('mesh_columns', [])
        coordinates = kwargs.get('mesh_coordinates', ['xyz', 'uvw'])
        mesh_datasets = kwargs.get('mesh_datasets', [])

        for t in times:
            if not isinstance(t, float):
                raise ValueError("times must all be of type float")


            # basic geometric columns
            if 'uvw' in coordinates:
                params += [FloatArrayParameter(qualifier='uvw_elements', time=t, value=kwargs.get('uvw_elements', []), readonly=True, default_unit=u.solRad, advanced=True, description='Vertices of triangles in the plane-of-sky')]
                params += [FloatArrayParameter(qualifier='uvw_normals', time=t, value=kwargs.get('uvw_normals', []), readonly=True, default_unit=u.solRad, advanced=True, description='Normals of triangles in the plane-of-sky')]

            if 'xyz' in coordinates:
                params += [FloatArrayParameter(qualifier='xyz_elements', time=t, value=kwargs.get('xyz_elements ', []), readonly=True, default_unit=u.dimensionless_unscaled, advanced=True, description='Vertices of triangles in Roche coordinates')]
                params += [FloatArrayParameter(qualifier='xyz_normals', time=t, value=kwargs.get('xyz_normals ', []), readonly=True, default_unit=u.dimensionless_unscaled, advanced=True, description='Normals of triangles in Roche coordinates')]

            # NOTE: if changing the parameters which are optional, changes must
            # be made here, in the choices for the columns Parameter, and in
            # backends.py when the values are extracted and included in the
            # packet

            # if 'pot' in columns:
                # params += [FloatParameter(qualifier='pot', time=t, value=kwargs.get('pot', 0.0), default_unit=u.dimensionless_unscaled, description='Equipotential of the stellar surface')]
            # if 'rpole' in columns:
            #     params += [FloatParameter(qualifier='rpole', time=t, value=kwargs.get('rpole', 0.0), default_unit=u.solRad, description='Polar radius of the stellar surface')]
            if 'volume' in columns:
                params += [FloatParameter(qualifier='volume', time=t, value=kwargs.get('volume', 0.0), readonly=True, default_unit=u.solRad**3, description='Volume of the stellar surface')]


            if 'xs' in columns:
                params += [FloatArrayParameter(qualifier='xs', time=t, value=kwargs.get('xs', []), readonly=True, default_unit=u.solRad, description='X coordinate of center of triangles in the plane-of-sky')]
            if 'ys' in columns:
                params += [FloatArrayParameter(qualifier='ys', time=t, value=kwargs.get('ys', []), readonly=True, default_unit=u.solRad, description='Y coordinate of center of triangles in the plane-of-sky')]
            if 'zs' in columns:
                params += [FloatArrayParameter(qualifier='zs', time=t, value=kwargs.get('zs', []), readonly=True, default_unit=u.solRad, description='Z coordinate of center of triangles in the plane-of-sky')]

            if 'vxs' in columns:
                params += [FloatArrayParameter(qualifier='vxs', time=t, value=kwargs.get('vxs', []), readonly=True, default_unit=u.km/u.s, description='X velocity of center of triangles')]
            if 'vys' in columns:
                params += [FloatArrayParameter(qualifier='vys', time=t, value=kwargs.get('vys', []), readonly=True, default_unit=u.km/u.s, description='Y velocity of center of triangles')]
            if 'vzs' in columns:
                params += [FloatArrayParameter(qualifier='vzs', time=t, value=kwargs.get('vzs', []), readonly=True, default_unit=u.km/u.s, description='Z velocity of center of triangles')]

            if 'nxs' in columns:
                params += [FloatArrayParameter(qualifier='nxs', time=t, value=kwargs.get('nxs', []), readonly=True, default_unit=u.dimensionless_unscaled, description='X component of normals')]
            if 'nys' in columns:
                params += [FloatArrayParameter(qualifier='nys', time=t, value=kwargs.get('nys', []), readonly=True, default_unit=u.dimensionless_unscaled, description='Y component of normals')]
            if 'nzs' in columns:
                params += [FloatArrayParameter(qualifier='nzs', time=t, value=kwargs.get('nzs', []), readonly=True, default_unit=u.dimensionless_unscaled, description='Z component of normals')]

            if 'us' in columns:
                params += [FloatArrayParameter(qualifier='us', time=t, value=kwargs.get('us', []), readonly=True, default_unit=u.solRad, description='U coordinate of center of triangles in the plane-of-sky')]
            if 'vs' in columns:
                params += [FloatArrayParameter(qualifier='vs', time=t, value=kwargs.get('vs', []), readonly=True, default_unit=u.solRad, description='V coordinate of center of triangles in the plane-of-sky')]
            if 'ws' in columns:
                params += [FloatArrayParameter(qualifier='ws', time=t, value=kwargs.get('ws', []), readonly=True, default_unit=u.solRad, description='W coordinate of center of triangles in the plane-of-sky')]

            if 'vus' in columns:
                params += [FloatArrayParameter(qualifier='vus', time=t, value=kwargs.get('vus', []), readonly=True, default_unit=u.km/u.s, description='U velocity of center of triangles')]
            if 'vvs' in columns:
                params += [FloatArrayParameter(qualifier='vvs', time=t, value=kwargs.get('vvs', []), readonly=True, default_unit=u.km/u.s, description='V velocity of center of triangles')]
            if 'vws' in columns:
                params += [FloatArrayParameter(qualifier='vws', time=t, value=kwargs.get('vws', []), readonly=True, default_unit=u.km/u.s, description='W velocity of center of triangles')]

            if 'nus' in columns:
                params += [FloatArrayParameter(qualifier='nus', time=t, value=kwargs.get('nus', []), readonly=True, default_unit=u.dimensionless_unscaled, description='U component of normals')]
            if 'nvs' in columns:
                params += [FloatArrayParameter(qualifier='nvs', time=t, value=kwargs.get('nvs', []), readonly=True, default_unit=u.dimensionless_unscaled, description='V component of normals')]
            if 'nws' in columns:
                params += [FloatArrayParameter(qualifier='nws', time=t, value=kwargs.get('nws', []), readonly=True, default_unit=u.dimensionless_unscaled, description='W component of normals')]


            if 'areas' in columns:
                params += [FloatArrayParameter(qualifier='areas', time=t, value=kwargs.get('areas', []), readonly=True, default_unit=u.solRad**2, description='Area of triangles')]
            # if 'tareas' in columns:
                # params += [FloatArrayParameter(qualifier='tareas', time=t, value=kwargs.get('areas', []), readonly=True, default_unit=u.solRad**2, description='Area of WD triangles')]


            if 'rs' in columns:
                params += [FloatArrayParameter(qualifier='rs', time=t, value=kwargs.get('rs', []), readonly=True, default_unit=u.solRad, description='Distance of each triangle from center of mass (of the half-envelope for contacts)')]
            # if 'cosbetas' in columns:
            #     params += [FloatArrayParameter(qualifier='cosbetas', time=t, value=kwargs.get('cosbetas', []), readonly=True, default_unit=u.solRad, description='')]


            if 'loggs' in columns:
                params += [FloatArrayParameter(qualifier='loggs', time=t, value=kwargs.get('loggs', []), readonly=True, default_unit=u.dimensionless_unscaled, description='Local surface gravity')]
            if 'teffs' in columns:
                params += [FloatArrayParameter(qualifier='teffs', time=t, value=kwargs.get('teffs', []), readonly=True, default_unit=u.K, description='Local effective temperature')]

            if 'rprojs' in columns:
                params += [FloatArrayParameter(qualifier='rprojs', time=t, value=kwargs.get('rprojs', []), readonly=True, default_unit=u.solRad, description='Projected distance (on plane of sky) of each triangle from center of mass (of the half-envelope for contacts)')]
            if 'mus' in columns:
                params += [FloatArrayParameter(qualifier='mus', time=t, value=kwargs.get('mus', []), readonly=True, default_unit=u.dimensionless_unscaled, description='Mu')]
            if 'visible_centroids' in columns:
                params += [FloatArrayParameter(qualifier='visible_centroids', time=t, value=kwargs.get('visible_centroids', []), readonly=True, default_unit=u.solRad  if t is not None else u.dimensionless_unscaled, description='Center of the visible portion of each triangle')]
            if 'visibilities' in columns:
                params += [FloatArrayParameter(qualifier='visibilities', time=t, value=kwargs.get('visibilities', []), readonly=True, default_unit=u.dimensionless_unscaled, description='Visiblity of triangles (1=visible, 0.5=partial, 0=hidden)')]

            # params += [FloatArrayParameter(qualifier='horizon_xs', time=t, value=kwargs.get('horizon_xs', []), readonly=True, default_unit=u.solRad, description='Horizon of the mesh (x component)')]
            # params += [FloatArrayParameter(qualifier='horizon_ys', time=t, value=kwargs.get('horizon_ys', []), readonly=True, default_unit=u.solRad, description='Horizon of the mesh (y component)')]
            # params += [FloatArrayParameter(qualifier='horizon_zs', time=t, value=kwargs.get('horizon_zs', []), readonly=True, default_unit=u.solRad, description='Horizon of the mesh (z component)')]
            # params += [FloatArrayParameter(qualifier='horizon_analytic_xs', time=t, value=kwargs.get('horizon_analytic_xs', []), readonly=True, default_unit=u.solRad, description='Analytic horizon (interpolated, x component)')]
            # params += [FloatArrayParameter(qualifier='horizon_analytic_ys', time=t, value=kwargs.get('horizon_analytic_ys', []), readonly=True, default_unit=u.solRad, description='Analytic horizon (interpolated, y component)')]
            # params += [FloatArrayParameter(qualifier='horizon_analytic_zs', time=t, value=kwargs.get('horizon_analytic_zs', []), readonly=True, default_unit=u.solRad, description='Analytic horizon (interpolated, z component)')]

            for dataset in mesh_datasets:
                # if 'dls@{}'.format(dataset) in columns:
                    # params += [FloatArrayParameter(qualifier='dls', dataset=dataset, time=t, value=[], default_unit=u.nm, description='Per-element delta-lambda caused by doppler shift'.format(dataset))]
                if 'rvs@{}'.format(dataset) in columns:
                    params += [FloatArrayParameter(qualifier='rvs', dataset=dataset, time=t, value=[], readonly=True, default_unit=u.km/u.s, description='Per-element value of rvs for {} dataset'.format(dataset))]
                if 'intensities@{}'.format(dataset) in columns:
                    params += [FloatArrayParameter(qualifier='intensities', dataset=dataset, time=t, value=[], readonly=True, default_unit=u.W/u.m**3, description='Per-element value of intensities for {} dataset'.format(dataset))]
                if 'normal_intensities@{}'.format(dataset) in columns:
                    params += [FloatArrayParameter(qualifier='normal_intensities', dataset=dataset, time=t, value=[], readonly=True, default_unit=u.W/u.m**3, description='Per-element value of normal_intensities for {} dataset'.format(dataset))]
                if 'abs_intensities@{}'.format(dataset) in columns:
                    params += [FloatArrayParameter(qualifier='abs_intensities', dataset=dataset, time=t, value=[], readonly=True, default_unit=u.W/u.m**3, description='Per-element value of abs_intensities for {} dataset'.format(dataset))]
                if 'abs_normal_intensities@{}'.format(dataset) in columns:
                    params += [FloatArrayParameter(qualifier='abs_normal_intensities', dataset=dataset, time=t, value=[], readonly=True, default_unit=u.W/u.m**3, description='Per-element value of abs_normal_intensities for {} dataset'.format(dataset))]
                if 'boost_factors@{}'.format(dataset) in columns:
                    params += [FloatArrayParameter(qualifier='boost_factors', dataset=dataset, time=t, value=[], readonly=True, default_unit=u.dimensionless_unscaled, description='Per-element value of boost_factors for {} dataset'.format(dataset))]
                if 'ldint@{}'.format(dataset) in columns:
                    params += [FloatArrayParameter(qualifier='ldint', dataset=dataset, time=t, value=kwargs.get('ldint', []), readonly=True, default_unit=u.dimensionless_unscaled, description='Integral of the limb-darkening function')]

                if 'ptfarea@{}'.format(dataset) in columns:
                    params += [FloatParameter(qualifier='ptfarea', dataset=dataset, time=t, value=kwargs.get('ptfarea', 1.0), readonly=True, default_unit=u.m, description='Area of the passband transmission function')]
                if 'pblum_ext@{}'.format(dataset) in columns:
                    params += [FloatParameter(qualifier='pblum_ext', dataset=dataset, time=t, value=kwargs.get('pblum_ext', 0.0), readonly=True, default_unit=u.W, description='Passband Luminosity of entire star (after pblum scaling)')]
                if 'abs_pblum_ext@{}'.format(dataset) in columns:
                    params += [FloatParameter(qualifier='abs_pblum_ext', dataset=dataset, time=t, value=kwargs.get('abs_pblum_ext', 0.0), readonly=True, default_unit=u.W, description='Passband Luminosity of entire star (before pblum scaling)')]


    return ParameterSet(params) if as_ps else params, constraints


# del _empty_array
# del deepcopy
# del download_passband, list_installed_passbands, list_online_passbands, list_passbands, parameter_from_json, parse_json, send_if_client, update_if_client
# del fnmatch



from . import nparray as _wrappers
import numpy as np
import json
import os

from distutils.version import LooseVersion

__version__ = '1.2.0'
version = __version__

# allow isinstance(obj, nparray.ndarray) to be similar to numpy
ndarray = _wrappers.ArrayWrapper

__docprefix__ = """
This is an nparray wrapper around the numpy function.  The
numpy documentation is included below.  Currently most kwargs
should be accepted with the exception of 'dtype'.  The returned
object should act exactly like the numpy array itself, but with
several extra helpful methods and attributes.  Call help on the
resulting object for more information.

If you have astropy installed, units are supported by passing unit=astropy.unit
to the instantiation functions or by multiplying an array with a unit object.

"""

__docsep__ = """

===============================================================

** numpy documentation for underlying function: **

"""

def array(value, unit=None):
    """
    Arguments
    ------------
    * `value` (array or list): array or list of values.
    * `unit` (astropy unit or string, optional, default=None): unit
      corresponding to the passed values.

    Returns
    -----------
    * <Array>
    """
    return _wrappers.Array(value, unit)

array.__doc__ = __docprefix__ + "\n".join([l.lstrip() for l in array.__doc__.split("\n")]) + __docsep__ + np.array.__doc__.replace("&gt;", ">")

def arange(start, stop, step, unit=None):
    """
    Arguments
    ------------
    * `start` (int or float): the starting point of the sequence.
    * `stop` (int or float): the ending point of the sequence.  The interval
        does not include this value, except in some cases where `step` is not an
        integer and floating point round-off affects the length of the array.
    * `step` (int or float): the stepsize between each item in the sequence.
    * `unit` (astropy unit or string, optional, default=None): unit
      corresponding to the passed values.

    Returns
    -----------
    * <Arange>
    """
    return _wrappers.Arange(start, stop, step, unit)

arange.__doc__ = __docprefix__ + "\n".join([l.lstrip() for l in arange.__doc__.split("\n")]) + __docsep__ + np.arange.__doc__.replace("&gt;", ">")

def linspace(start, stop, num, endpoint=True, unit=None):
    """
    Arguments
    ------------
    * `start` (int or float): the starting point of the sequence.
    * `stop` (int or float): the ending point of the sequence, unless `endpoint`
        is set to False.  In that case, the sequence consists of all but the
        last of ``num + 1`` evenly spaced samples, so that `stop` is excluded.
        Note that the step size changes when `endpoint` is False.
    * `num` (int): number of samples to generate.
    * `endpoint` (bool, optional, default=True): If True, `stop` is the last
        sample. Otherwise, it is not included.
    * `unit` (astropy unit or string, optional, default=None): unit
        corresponding to the passed values.

    Returns
    -----------
    * <Linspace>
    """
    return _wrappers.Linspace(start, stop, num, endpoint, unit)

linspace.__doc__ = __docprefix__ + "\n".join([l.lstrip() for l in linspace.__doc__.split("\n")]) + __docsep__ + np.linspace.__doc__.replace("&gt;", ">")

def invspace(start, stop, num, endpoint=True, unit=None):
    """
    Evenly sampled numbers in inverted space.  This is equivalent to:

    ```py
    1./linspace(1./start, 1./stop, num, endpoint, unit)
    ```

    See also:

    * <nparray.linspace>

    Arguments
    ------------
    * `start` (int or float): the starting point of the sequence.
    * `stop` (int or float): the ending point of the sequence, unless `endpoint`
        is set to False.  In that case, the sequence consists of all but the
        last of ``num + 1`` evenly spaced samples, so that `stop` is excluded.
        Note that the step size changes when `endpoint` is False.
    * `num` (int): number of samples to generate.
    * `endpoint` (bool, optional, default=True): If True, `stop` is the last
        sample. Otherwise, it is not included.
    * `unit` (astropy unit or string, optional, default=None): unit
        corresponding to the passed values.

    Returns
    -----------
    * <Invspace>
    """
    return _wrappers.Invspace(start, stop, num, endpoint, unit)

def logspace(start, stop, num, endpoint=True, base=10.0, unit=None):
    """
    See also:

    * <nparray.geomspace>

    Arguments
    ------------
    * `start` (int or float): ``base ** start`` is the starting value of the sequence.
    * `stop` (int or float): ``base ** stop`` is the final value of the sequence,
        unless `endpoint` is False.  In that case, ``num + 1`` values are spaced
        over the interval in log-space, of which all but the last (a sequence of
        length `num`) are returned.
    * `num` (int): number of samples to generate.
    * `endpoint` (bool, optional, default=True): If True, `stop` is the last
        sample. Otherwise, it is not included.
    * `base` (float, optional, default=10.0): The base of the log space. The
        step size between the elements in ``ln(samples) / ln(base)``
        (or ``log_base(samples)``) is uniform.
    * `unit` (astropy unit or string, optional, default=None): unit
      corresponding to the passed values.

    Returns
    -----------
    * <Logspace>
    """
    return _wrappers.Logspace(start, stop, num, endpoint, base, unit)

logspace.__doc__ = __docprefix__ + "\n".join([l.lstrip() for l in logspace.__doc__.split("\n")]) + __docsep__ + np.logspace.__doc__.replace("&gt;", ">")

def geomspace(start, stop, num, endpoint=True, unit=None):
    """
    See also:

    * <nparray.logspace>

    Arguments
    ------------
    * `start` (int or float): the starting point of the sequence.
    * `stop` (int or float): the final value of the sequence, unless `endpoint`
        is False.  In that case, ``num + 1`` values are spaced over the
        interval in log-space, of which all but the last (a sequence of
        length `num`) are returned.
    * `num` (int): number of samples to generate.
    * `endpoint` (bool, optional, default=True): If True, `stop` is the last
        sample. Otherwise, it is not included.
    * `unit` (astropy unit or string, optional, default=None): unit
      corresponding to the passed values.

    Returns
    -----------
    * <Geomspace>
    """
    if LooseVersion(np.__version__) >= LooseVersion("1.13"):
        return _wrappers.Geomspace(start, stop, num, endpoint, unit)
    else:
        raise NotImplementedError("geomspace requires numpy version >= 1.13")

if LooseVersion(np.__version__) >= LooseVersion("1.13"):
    geomspace.__doc__ = __docprefix__ + "\n".join([l.lstrip() for l in geomspace.__doc__.split("\n")]) + __docsep__ + np.geomspace.__doc__.replace("&gt;", ">")

def full(shape, fill_value, unit=None):
    """
    See also:

    * <nparray.full_like>

    Arguments
    ------------
    * `shape` (int or sequence of ints): Shape of the new array, e.g.,
        ``(2, 3)`` or ``2``.
    * `fill_value` (int or float): Value to fill each element in the array.
    * `unit` (astropy unit or string, optional, default=None): unit
      corresponding to the passed values.

    Returns
    -----------
    * <Full>
    """
    return _wrappers.Full(shape, fill_value, unit)

full.__doc__ = __docprefix__ + "\n".join([l.lstrip() for l in full.__doc__.split("\n")]) + __docsep__ + np.full.__doc__.replace("&gt;", ">")

def full_like(a, fill_value, unit=None):
    """
    Note: unlike in the numpy version, the data-type of `a` is not currently
    guaranteed to be maintained.

    See also:

    * <nparray.full>

    Arguments
    ------------
    * `a` (list or array): The shape of `a` define these same attributes of the
        returned array.
    * `fill_value` (int or float): Value to fill each element in the array.
    * `unit` (astropy unit or string, optional, default=None): unit
      corresponding to the passed values.

    Returns
    -----------
    * <Full>
    """
    return _wrappers.Full(a.shape, fill_value, unit)

full_like.__doc__ = __docprefix__ + "\n".join([l.lstrip() for l in full_like.__doc__.split("\n")]) + __docsep__ + np.full_like.__doc__.replace("&gt;", ">")

def zeros(shape, unit=None):
    """
    See also:

    * <nparray.zeros_like>

    Arguments
    ------------
    * `shape` (int or sequence of ints): Shape of the new array, e.g.,
        ``(2, 3)`` or ``2``.
    * `unit` (astropy unit or string, optional, default=None): unit
      corresponding to the passed values.

    Returns
    -----------
    * <Zeros>
    """
    return _wrappers.Zeros(shape, unit)

zeros.__doc__ = __docprefix__ + "\n".join([l.lstrip() for l in zeros.__doc__.split("\n")]) + __docsep__ + np.zeros.__doc__.replace("&gt;", ">")

def zeros_like(a, unit=None):
    """
    Note: unlike in the numpy version, the data-type of `a` is not currently
    guaranteed to be maintained.

    See also:

    * <nparray.zeros>

    Arguments
    ------------
    * `a` (list or array): The shape of `a` define these same attributes of the
        returned array.
    * `unit` (astropy unit or string, optional, default=None): unit
      corresponding to the passed values.

    Returns
    -----------
    * <Zeros>
    """
    return _wrappers.Zeros(a.shape, unit)

zeros_like.__doc__ = __docprefix__ + "\n".join([l.lstrip() for l in zeros_like.__doc__.split("\n")]) + __docsep__ + np.zeros_like.__doc__.replace("&gt;", ">")

def ones(shape, unit=None):
    """
    See also:

    * <nparray.ones_like>

    Arguments
    ------------
    * `shape` (int or sequence of ints): Shape of the new array, e.g.,
        ``(2, 3)`` or ``2``.
    * `unit` (astropy unit or string, optional, default=None): unit
      corresponding to the passed values.

    Returns
    -----------
    * <Ones>
    """
    return _wrappers.Ones(shape, unit)

ones.__doc__ = __docprefix__ + "\n".join([l.lstrip() for l in ones.__doc__.split("\n")]) + __docsep__ + np.ones.__doc__.replace("&gt;", ">")

def ones_like(a, unit=None):
    """
    Note: unlike in the numpy version, the data-type of `a` is not currently
    guaranteed to be maintained.

    See also:

    * <nparray.ones>

    Arguments
    ------------
    * `a` (list or array): The shape of `a` define these same attributes of the
        returned array.
    * `unit` (astropy unit or string, optional, default=None): unit
      corresponding to the passed values.

    Returns
    -----------
    * <Ones>
    """
    return _wrappers.Ones(a.shape, unit)

ones_like.__doc__ = __docprefix__ + "\n".join([l.lstrip() for l in ones_like.__doc__.split("\n")]) + __docsep__ + np.ones_like.__doc__.replace("&gt;", ">")

def eye(M, N=None, k=0, unit=None):
    """
    Arguments
    ------------
    * `M` (int): Number of rows in the output.
    * `N` (int or None, optional, default=None): Number of columns in the output.
        If None, defaults to `N`.
    * `k` (int, optional, default=0): Index of the diagonal: 0 (the default)
        refers to the main diagonal, a positive value refers to an upper
        diagonal, and a negative value to a lower diagonal.
    * `unit` (astropy unit or string, optional, default=None): unit
      corresponding to the passed values.

    Returns
    -----------
    * <Eye>
    """
    return _wrappers.Eye(M, N, k, unit)

eye.__doc__ = __docprefix__ + "\n".join([l.lstrip() for l in eye.__doc__.split("\n")]) + __docsep__ + np.eye.__doc__.replace("&gt;", ">")

def from_dict(d):
    """
    Load an nparray object from a dictionary.

    See also:

    * <nparray.from_json>
    * <nparray.from_file>

    Arguments
    ----------
    * `d` (dictionary): dictionary representing a valid nparray object

    Returns
    ---------
    * the instatiated nparray object.
    """
    if isinstance(d, str):
        return from_json(d)

    if not isinstance(d, dict):
        raise TypeError("argument must be of type dict")
    if 'nparray' not in d.keys():
        raise ValueError("input dictionary missing 'nparray' entry")

    classname = d.get('nparray').title()
    # instead of popping npdarray (which would happen in memory and make that json
    # unloadable again), we'll do a dictionary comprehension.  If this causes
    # performance issues, we could instead accept and ignore nparray as
    # a keyword argument to __init__
    return getattr(_wrappers, classname)(**{k:v for k,v in d.items() if k!='nparray'})

def from_json(j):
    """
    Load an nparray object from a json-formatted string.

    See also:

    * <nparray.from_dict>
    * <nparray.from_file>

    Arguments
    -----------
    * `j` (string): a json-formatted string representing a valid nparray object

    Returns
    ----------
    * the instantiated nparray object.
    """
    if isinstance(j, dict):
        return from_dict(j)

    if not (isinstance(j, str) or isinstance(j, unicode)):
        raise TypeError("argument must be of type str")

    return from_dict(json.loads(j))

def from_file(filename):
    """
    Load an nparray object from a file.

    See also:

    * <nparray.from_dict>
    * <nparray.from_json>

    Arguments
    -----------
    * `filename` (string): a filename pointing to a valid nparray object.

    Returns
    ----------
    * the instantiated nparray object.
    """
    f = open(filename, 'r')
    j = json.load(f)
    f.close()

    return from_dict(j)

def monkeypatch():
    """
    monkeypatch built-in numpy functions to call those provided by nparray instead.

    ```py
    import nparray
    import numpy as np

    nparray.monkeypatch()
    print(np.linspace(0,1,11))
    ```

    """
    np.array = array
    np.arange = arange
    np.linspace = linspace
    np.logspace = logspace
    np.geomspace = geomspace
    np.full = full
    np.full_like = full_like
    np.zeros = zeros
    np.zeros_like = zeros_like
    np.ones = ones
    np.ones_like = ones_like
    np.eye = eye





from phoebe.parameters import *
from phoebe.parameters import constraint
from phoebe import u
from phoebe import conf

### NOTE: if creating new parameters, add to the _forbidden_labels list in parameters.py

def _component_allowed_for_feature(feature_kind, component_kind):
    _allowed = {}
    _allowed['spot'] = ['star', 'envelope']
    _allowed['pulsation'] = ['star', 'envelope']
    _allowed['gaussian_process'] = [None]

    return component_kind in _allowed[feature_kind]

def _dataset_allowed_for_feature(feature_kind, dataset_kind):
    _allowed = {}
    _allowed['spot'] = [None]
    _allowed['pulsation'] = [None]
    _allowed['gaussian_process'] = ['lc', 'rv', 'lp']

    return dataset_kind in _allowed[feature_kind]

def spot(feature, **kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for a spot feature.

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_feature>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_feature>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    Allowed to attach to:
    * components with kind: star
    * datasets: not allowed

    Arguments
    ----------
    * `colat` (float/quantity, optional): colatitude of the center of the spot
        wrt spin axis.
    * `long` (float/quantity, optional): longitude of the center of the spot wrt
        spin axis.
    * `radius` (float/quantity, optional): angular radius of the spot.
    * `relteff` (float/quantity, optional): temperature of the spot relative
        to the intrinsic temperature.

    Returns
    --------
    * (<phoebe.parameters.ParameterSet>, list): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects and a list of all necessary
        constraints.
    """

    params = []

    params += [FloatParameter(qualifier="colat", value=kwargs.get('colat', 0.0), default_unit=u.deg, description='Colatitude of the center of the spot wrt spin axis')]
    params += [FloatParameter(qualifier="long", value=kwargs.get('long', 0.0), default_unit=u.deg, description='Longitude of the center of the spot wrt spin axis')]
    params += [FloatParameter(qualifier='radius', value=kwargs.get('radius', 1.0), default_unit=u.deg, description='Angular radius of the spot')]
    # params += [FloatParameter(qualifier='area', value=kwargs.get('area', 1.0), default_unit=u.solRad, description='Surface area of the spot')]

    params += [FloatParameter(qualifier='relteff', value=kwargs.get('relteff', 1.0), limits=(0.,None), default_unit=u.dimensionless_unscaled, description='Temperature of the spot relative to the intrinsic temperature')]
    # params += [FloatParameter(qualifier='teff', value=kwargs.get('teff', 10000), default_unit=u.K, description='Temperature of the spot')]

    constraints = []

    return ParameterSet(params), constraints

def gaussian_process(feature, **kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for a gaussian_process feature.

    Requires celerite to be installed.  See https://celerite.readthedocs.io/en/stable/.
    If using gaussian processes, consider citing:
    * https://ui.adsabs.harvard.edu/abs/2017AJ....154..220F

    See also:
    * <phoebe.frontend.bundle.Bundle.references>

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_feature>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_feature>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    Allowed to attach to:
    * components: not allowed
    * datasets with kind: lc

    If `compute_times` or `compute_phases` is used: the underlying model without
    gaussian_processes will be computed at the given times/phases but will then
    be interpolated into the times of the underlying dataset to include the
    contribution of gaussian processes and will be exposed at the dataset
    times (with a warning in the logger and in
    <phoebe.frontend.bundle.Bundle.run_checks_compute>).  If the system is
    time-dependent without GPs
    (see <phoebe.parameters.HierarchyParameter.is_time_dependent>), then
    the underlying model will need to cover the entire dataset or an error
    will be raised by <phoebe.frontend.bundle.Bundle.run_checks_compute>.


    Arguments
    ----------
    * `kernel` (string, optional, default='matern32'): Kernel for the gaussian
        process (see https://celerite.readthedocs.io/en/stable/python/kernel/)
    * `log_S0` (float, optional, default=0): only applicable if `kernel` is
        'sho'. Log of the GP parameter S0.
    * `log_Q` (float, optional, default=0): only applicable if `kernel` is
        'sho'.  Log of the GP parameter Q.
    * `log_omega0` (float, optional, default=0): only applicable if `kernel` is
        'sho'.  Log of the GP parameter omega0.
    * `log_sigma` (float, optional, default=0): only applicable if `kernel` is
        'matern32'.  Log of the GP parameter sigma.
    * `log_rho` (float, optional, default=0): only applicable if `kernel` is
        'matern32'.  Log of the GP parameter rho.
    * `eps` (float, optional, default=0): only applicable if `kernel` is
        'matern32'.  Log of the GP parameter epsilon.


    Returns
    --------
    * (<phoebe.parameters.ParameterSet>, list): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects and a list of all necessary
        constraints.
    """

    params = []

    params += [ChoiceParameter(qualifier='kernel', value=kwargs.get('kernel', 'matern32'), choices=['matern32', 'sho'], description='Kernel for the gaussian process (see https://celerite.readthedocs.io/en/stable/python/kernel/)')]

    params += [FloatParameter(visible_if='kernel:sho', qualifier='log_S0', value=kwargs.get('log_S0', 0), default_unit=u.dimensionless_unscaled, description='Log of the GP parameter S0')]
    params += [FloatParameter(visible_if='kernel:sho', qualifier='log_Q', value=kwargs.get('log_Q', 0), default_unit=u.dimensionless_unscaled, description='Log of the GP parameter Q')]
    params += [FloatParameter(visible_if='kernel:sho', qualifier='log_omega0', value=kwargs.get('log_omega0', 0), default_unit=u.dimensionless_unscaled, description='Log of the GP parameter omega0')]

    params += [FloatParameter(visible_if='kernel:matern32', qualifier='log_sigma', value=kwargs.get('log_sigma', 0), default_unit=u.dimensionless_unscaled, description='Log of the GP parameter sigma')]
    params += [FloatParameter(visible_if='kernel:matern32', qualifier='log_rho', value=kwargs.get('log_rho', 0), default_unit=u.dimensionless_unscaled, description='Log of the GP parameter rho')]
    params += [FloatParameter(visible_if='kernel:matern32', qualifier='eps', value=kwargs.get('eps', 0.01), limits=(0,None), default_unit=u.dimensionless_unscaled, description='GP parameter epsilon')]

    # params += [FloatParameter(visible_if='kernel:jitter', qualifier='log_sigma', value=kwargs.get('log_sigma', np.log(0.01)), default_unit=u.dimensionless_unscaled, description='Log of the amplitude of the white noise')]

    constraints = []

    return ParameterSet(params), constraints



# del deepcopy
# del _component_allowed_for_feature
# del download_passband, list_installed_passbands, list_online_passbands, list_passbands, parameter_from_json, parse_json, send_if_client, update_if_client
# del fnmatch



"""
Non-standard interpolation methods.
"""
import time
import itertools
import libphoebe   
import numpy as np
from scipy import ndimage
from scipy.ndimage import _nd_image, _ni_support
try:
    import pyfits as pf
except:
    try: # Pyfits now integrated in astropy
        import astropy.io.fits as pf
    except ImportError:
        print(("Soft warning: pyfits could not be found on your system, you can "
           "only use black body atmospheres and Gaussian synthetic spectrallll "
           "profiles"))


def create_pixeltypegrid(grid_pars, grid_data):
    """
    Creates pixelgrid and arrays of axis values.
    
    Starting from:
    
        - grid_pars: 2D numpy array, 1 column per parameter, unlimited number of
          cols
        - grid_data: 2D numpy array, 1 column per variable, data corresponding
          to the rows in grid_pars
    
    
    The grid should be rectangular and complete, i.e. every combination of the unique values in the 
    parameter columns should exist. If not, a nan value will be inserted.
    
    @param grid_pars: Npar x Ngrid array of parameters
    @type grid_pars: array
    @param grid_data: Ndata x Ngrid array of data
    @type grid_data: array
    @return: axis values and pixelgrid
    @rtype: array, array
    """

    uniques = [np.unique(column, return_inverse=True) for column in grid_pars]
    #[0] are the unique values, [1] the indices for these to recreate the original array
    
    # we need to copy the values of the unique axes explicitly into new arrays
    # otherwise we can get issues with the interpolator
    axis_values = []
    for uniques_ in uniques:
        this_axes = np.zeros(len(uniques_[0]))
        this_axes[:] = uniques_[0]
        axis_values.append(this_axes)
    #axis_values = [uniques_[0] for uniques_ in uniques]
    #axis_values = [np.require(uniques_[0],requirements=['A','O','W','F']) for uniques_ in uniques]
    
    unique_val_indices = [uniques_[1] for uniques_ in uniques]
    
    data_dim = np.shape(grid_data)[0]

    par_dims   = [len(uv[0]) for uv in uniques]

    par_dims.append(data_dim)
    pixelgrid = np.ones(par_dims)
    
    # We put np.inf as default value. If we get an inf, that means we tried to access
    # a region of the pixelgrid that is not populated by the data table
    pixelgrid[pixelgrid==1] = np.inf
    
    # now populate the multiDgrid
    indices = [uv[1] for uv in uniques]
    pixelgrid[indices] = grid_data.T
    return tuple(axis_values), pixelgrid


def interpolate(p, axis_values, pixelgrid, order=1, mode='constant', cval=0.0):
    """
    Interpolates in a grid prepared by create_pixeltypegrid().
    
    p is an array of parameter arrays
    
    @param p: Npar x Ninterpolate array
    @type p: array
    @return: Ndata x Ninterpolate array
    @rtype: array
    """
    # convert requested parameter combination into a coordinate
    p_ = np.array([np.searchsorted(av_,val) for av_, val in zip(axis_values,p)])
    lowervals_stepsize = np.array([[av_[p__-1], av_[p__]-av_[p__-1]] \
                         for av_, p__ in zip(axis_values,p_)])
    p_coord = (p-lowervals_stepsize[:,0])/lowervals_stepsize[:,1] + p_-1
    
    # interpolate
    if order > 1:
        prefilter = False
    else:
        prefilter = False

    return [ndimage.map_coordinates(pixelgrid[...,i],p_coord, order=order,
                                    prefilter=prefilter, mode=mode, cval=cval) \
                for i in range(np.shape(pixelgrid)[-1])]


def cinterpolate(p, axis_values, pixelgrid):
    """
    Interpolates in a grid prepared by create_pixeltypegrid().
    
    Does a similar thing as :py:func:`interpolate`, but does everything in C.
    
    p is an array of parameter arrays.
    
    Careful, the shape of input :envvar:`p` and output is the transpose of
    :py:func:`interpolate`.
    
    @param p: Ninterpolate X Npar array
    @type p: array
    @return: Ninterpolate X Ndata array
    @rtype: array
    """
    res = libphoebe.interp(p, axis_values, pixelgrid)
    return res


if __name__ == "__main__":
    
    # mock data
    x = np.linspace(0,10,20)
    y = np.linspace(0,10,30)
    z = np.linspace(0,10,30)
    a = np.linspace(0,10,30)
    grid_pars = []
    
    for pars in itertools.product(x,y,z,a):
        grid_pars.append(pars)    
    grid_pars = np.array(grid_pars).T
    grid_data = np.array([1000000*grid_pars[0]+10000*grid_pars[1]+100*grid_pars[2]+grid_pars[3],100000*grid_pars[0]+1000*grid_pars[1]+10*grid_pars[2]+grid_pars[3]/10.,1000000*grid_pars[0]+10000*grid_pars[1]+100*grid_pars[2]+grid_pars[3],1000000*grid_pars[0]+10000*grid_pars[1]+100*grid_pars[2]+grid_pars[3]])
    print(np.shape(grid_data))
    p = np.array([[1]*10000,[1.2]*10000,[3.1]*10000,[0.1]*10000])
    print("Creating array")
    c0=time.time()    
    axis_values, pixelgrid = create_pixeltypegrid(grid_pars, grid_data)
    c1= time.time()
    print("Interpolating")
    vals = interpolate(p, axis_values,pixelgrid)
    
    c2=time.time()
    
    print("Creating array:", c1-c0, 's')
    print("Interpolation:", c2-c1, 's')
    print(vals)
    # ---------------------------------------------------------------------



"""
>>> import phoebe

Available environment variables:
* PHOEBE_ENABLE_PLOTTING=TRUE/FALSE (whether to import plotting libraries with phoebe: defaults to True)
* PHOEBE_ENABLE_SYMPY=TRUE/FALSE (whether to attempt to import sympy for constraint algebra: defaults to True if sympy installed, otherwise False)
* PHOEBE_ENABLE_ONLINE_PASSBANDS=TRUE/FALSE (whether to query for online passbands and download on-the-fly: defaults to True)
* PHOEBE_PBDIR (directory to search for passbands, in addition to phoebe.list_passband_directories())
* PHOEBE_DOWNLOAD_PASSBAND_DEFAULTS_GZIPPED=TRUE/FALSE (whether to download gzipped version of passbands by default.  Defaults to False.  Note that gzipped files take longer to load and will increase time for import, but take significantly less disk-space.)
* PHOEBE_DOWNLOAD_PASSBAND_DEFAULTS_CONTENT (default content, comma separated for list.  Defaults to 'all')
* PHOEBE_UPDATE_PASSBAND_IGNORE_VERSION=TRUE/FALSE (update passbands that need new content even if the online version is newer than the installed version.  Defaults to False.)
* PHOEBE_ENABLE_MPI=TRUE/FALSE (whether to use internal parallelization: defaults to True if within mpirun, otherwise False, can override in python with phoebe.mpi.on() and phoebe.mpi.off())
* PHOEBE_MPI_NPROCS=INT (number of procs to spawn in mpi is enabled but not running within mpirun: defaults to 4, only applicable if not within mpirun and PHOEBE_ENABLE_MPI=TRUE or phoebe.mpi.on() called, can override in python by passing nprocs to phoebe.mpi.on() or by setting phoebe.mpi.nprocs)
* PHOEBE_MULTIPROC_NPROCS=INT (number of proces to use within multiprocessing.  Multiprocessing is used for solver that support it and when sampling over a distribution in run_compute if MPI is not in use.  Set to 0 to disable multiprocessing and force serial.  Defaults to number of CPUs available.)
* PHOEBE_PBDIR (directory to search for passbands, in addition to phoebe.list_passband_directories())
* PHOEBE_DEVEL=TRUE/FALSE enable developer mode by default

"""

__version__ = '2.3.42'

import os as _os
import sys as _sys
import inspect as _inspect
import multiprocessing as _multiprocessing
import atexit
import re

# People shouldn't import Phoebe from the installation directory (inspired upon
# pymc warning message).
if _os.getcwd().find(_os.path.abspath(_os.path.split(_os.path.split(__file__)[0])[0]))>-1:
    # We have a clash of package name with the standard library: we implement an
    # "io" module and also they do. This means that you can import Phoebe from its
    # main source tree; then there is no difference between io from here and io
    # from the standard library. Thus, if the user loads the package from here
    # it will never work. Instead of letting Python raise the io clash (which
    # is uniformative to the unexperienced user), we raise the importError here
    # with a helpful error message
    raise ImportError('\n\tYou cannot import Phoebe from inside its main source tree.\n')

def _env_variable_string_or_list(key, default):
    value = _os.getenv(key, default)
    if "," in value:
        return value.split(",")
    else:
        return value

def _env_variable_int(key, default):
    value = _os.getenv(key, default)

    return int(value)

def _env_variable_int_or_none(key, default):
    value = _os.getenv(key, default)
    if value is None or isinstance(value, str) and value.lower()=='none':
        return None
    if isinstance(value, str):
        value = float(value)
    return int(value)

def _env_variable_bool(key, default):
    value = _os.getenv(key, default)
    if isinstance(value, bool):
        return value
    elif value.upper()=='TRUE':
        return True
    else:
        return False

# If we try to load matplotlib.pyplot on a non-X system, it will fail
# unless 'Agg' is used before the import. All X-systems define the
# 'DISPLAY' environment variable, and all non-X-systems do not. We do make a
# distinction between windows and unix based system. Hence:
if _env_variable_bool('PHOEBE_ENABLE_PLOTTING', True):
    try:
        import matplotlib
    except ImportError:
        pass
        # we'll catch this later in plotting and throw warnings as necessary
    else:
        if 'DISPLAY' not in _os.environ.keys() and _sys.platform not in ['win32','cygwin']:
            matplotlib.use('Agg')
        elif hasattr(_sys, 'real_prefix'):
            # then we're likely in a virtualenv.  Our best bet is to use the 'TkAgg'
            # backend, but this will require python-tk to be installed on the system
            try:
                matplotlib.use('Agg')
            except:
                matplotlib.use('TkAgg')



import logging
_logger = logging.getLogger("PHOEBE")
_logger.addHandler(logging.NullHandler())

###############################################################################
#########################         BEGIN MPI          ##########################
###############################################################################

# detect if we're within mpirun and if so, place all non-zero-rank
# processors into a wait loop.  This must happen before we start importing from
# phoebe so that those can have access to the _mpi object.

class MPI(object):
    def __init__(self):
        # this is a bit of a hack and will only work with openmpi, but environment
        # variables seem to be the only way to detect whether the script was run
        # via mpirun or not
        evars = _os.environ.keys()
        if 'OMPI_COMM_WORLD_SIZE' in evars or 'MV2_COMM_WORLD_SIZE' in evars or 'PMI_SIZE' in evars:
            from mpi4py import MPI as mpi4py
            self._within_mpirun = True
            self._internal_mpi = True

            self._comm   = mpi4py.COMM_WORLD
            self._myrank = self.comm.Get_rank()
            self._nprocs = self.comm.Get_size()

            if self._nprocs==1:
                raise ImportError("need more than 1 processor to run with mpi")

            self._enabled = _env_variable_bool("PHOEBE_ENABLE_MPI", True)

        else:
            self._within_mpirun = False
            self._internal_mpi = False
            self._comm = None
            self._myrank = 0
            self._nprocs = _env_variable_int("PHOEBE_MPI_NPROCS", 4)

            self._enabled = _env_variable_bool("PHOEBE_ENABLE_MPI", False)


    def __repr__(self):
        return "<MPI mode={} myrank={} nprocs={}>".format(self.mode, self.myrank, self.nprocs)

    @property
    def mode(self):
        if self.within_mpirun:
            if self.enabled:
                return "internal handling of mpi within mpirun"
            else:
                return "external handling of mpi by the user within mpirun"
        else:
            if self.enabled:
                return "internal handling of mpi in spawned separate threads during run_compute"
            else:
                return "serial mode"

    @property
    def enabled(self):
        return self._enabled

    def on(self, nprocs=None):
        if self.within_mpirun and not self.enabled:
            raise ValueError("cannot enable mpi after disabling within mpirun.")

        self._enabled = True

        if nprocs is not None:
            self.nprocs = nprocs

    def off(self):
        if self.within_mpirun and self.myrank == 0:
            self.comm.bcast({'worker_command': 'release'}, root=0)

        self._enabled = False

    @property
    def myrank(self):
        return self._myrank

    @property
    def nprocs(self):
        if not self.enabled and not self.within_mpirun:
            return 1
        else:
            return self._nprocs

    @nprocs.setter
    def nprocs(self, nprocs):
        if self.within_mpirun:
            _logger.warning("ignoring setting nprocs while within mpirun, nprocs={}".format(self.nprocs))
        else:
            self._nprocs = nprocs

    @property
    def comm(self):
        return self._comm

    @property
    def within_mpirun(self):
        return self._within_mpirun

    @property
    def detach_cmd(self):
        if self.within_mpirun:
            raise ValueError("detach not available within mpirun")

        # TODO: allow this as an option in the settings?
        python = 'python3'

        if self.enabled:
            return 'mpiexec -np %d %s {}' % (self.nprocs, python)
        else:
            return '%s {}' % python

    def shutdown_workers(self):
        if self.within_mpirun and self.myrank == 0:
            self.comm.bcast({'worker_command': 'shutdown'}, root=0)
            self._enabled = False
            # even though technically not true, we're now strictly serial and have no way of regaining the workers
            self._within_mpirun = False


mpi = MPI()

# NOTE: logic for worker waiting for tasks below after phoebe imports

###############################################################################
##########################         END MPI          ###########################
###############################################################################


###############################################################################
#########################        BEGIN SETTINGS        ########################
###############################################################################

class Settings(object):
    def __init__(self):
        # Check to see whether in interactive mode
        import __main__
        # hasattr(__main__, '__file__') will be True if running a python script, but
        # false if in a python or ipython interpreter.
        # _sys.flags.interactive will be 1 if the -i flag is sent to python

        # For now we'll set interactive_constraints to True by default, requiring it to
        # explicitly be disabled.
        # See #154 (https://github.com/phoebe-project/phoebe2/issues/154)
        self._interactive_constraints = True

        # We'll set interactive system checks to be on if running within a Python
        # console, but False if running from within a script
        # See #255 (https://github.com/phoebe-project/phoebe2/issues/255)
        self._interactive_checks = not hasattr(__main__, '__file__') or bool(_sys.flags.interactive)

        self._download_passband_defaults = {'content': _env_variable_string_or_list('PHOEBE_DOWNLOAD_PASSBAND_DEFAULTS_CONTENT', 'all'),
                                            'gzipped': _env_variable_bool('PHOEBE_DOWNLOAD_PASSBAND_DEFAULTS_GZIPPED', False)}

        self._update_passband_ignore_version = _env_variable_bool('PHOEBE_UPDATE_PASSBAND_IGNORE_VERSION', False)

        self._multiprocessing_nprocs = _env_variable_int_or_none('PHOEBE_MULTIPROC_NPROCS', None)
        self._progressbars = True

        # And we'll require explicitly setting developer mode on
        self._devel = _env_variable_bool('PHOEBE_DEVEL', False)

    def __repr__(self):
        return "<Settings interactive_checks={} interactive_constraints={}>".format(self.interactive_checks, self.interactive_constraints)

    def reset(self):
        self.__init__()

    def interactive_on(self):
        self.interactive_checks_on()
        self.interactive_constraints_on()

    def interactive_off(self, suppress_warning=False):
        self.interactive_checks_off(suppress_warning=suppress_warning)
        self.interactive_constraints_off(suppress_warning=suppress_warning)

    def interactive_checks_on(self):
        self._interactive_checks = True

    def interactive_checks_off(self, suppress_warning=False):
        if not suppress_warning:
            _logger.warning("checks will not be run until 'run_checks' or 'run_compute' is called.")
        self._interactive_checks = False

    def interactive_constraints_on(self):
        self._interactive_constraints = True

    def interactive_constraints_off(self, suppress_warning=False):
        if not suppress_warning:
            _logger.warning("constraints will not be run until 'run_delayed_constraints' or 'run_compute' is called.  This may result in inconsistent parameters if printing values before calling either of these methods.")
        self._interactive_constraints = False

    @property
    def interactive_checks(self):
        return self._interactive_checks

    @property
    def interactive_constraints(self):
        return self._interactive_constraints

    def devel_on(self):
        self._devel = True

    def devel_off(self):
        self._devel = False

    @property
    def devel(self):
        return self._devel

    def set_download_passband_defaults(self, **kwargs):
        """
        """
        for k,v in kwargs.items():
            if k not in self._download_passband_defaults.keys():
                raise KeyError("{} must be one of {}".format(self._download_passband_defaults.keys()))
            if k=='content' and not (isinstance(v, str) or isinstance(v, list)):
                raise TypeError("content must be of type string or list")
            if k=='gzipped' and not (isinstance(v, bool)):
                raise TypeError("gzipped must be of type bool")
            self._download_passband_defaults[k] = v

    def get_download_passband_defaults(self):
        return self._download_passband_defaults

    @property
    def download_passband_defaults(self):
        return self._download_passband_defaults

    @property
    def update_passband_ignore_version(self):
        return self._update_passband_ignore_version

    def update_passband_ignore_version_on(self):
        self._update_passband_ignore_version = True

    def update_passband_ignore_version_on(self):
        self._update_passband_ignore_version = False

    def multiprocessing_off(self):
        self._multiprocessing_nprocs = 0

    def multiprocessing_on(self):
        self._multiprocessing_nprocs = None

    def multiprocessing_set_nprocs(self, value):
        if not isinstance(value, int):
            return TypeError("must be integer")
        if value > _multiprocessing.cpu_count():
            return ValueError("only {} CPUs available".format(value))
        elif value < 0:
            return ValueError("nprocs must be >= 0")
        self._multiprocessing_nprocs = value

    @property
    def multiprocessing_nprocs(self):
        if self._multiprocessing_nprocs is None:
            return _multiprocessing.cpu_count()
        return self._multiprocessing_nprocs

    def progressbars_on(self):
        self._progressbars = True

    def progressbars_off(self):
        self._progressbars = False

    @property
    def progressbars(self):
        return self._progressbars

conf = Settings()

###############################################################################
##########################        END SETTINGS        #########################
###############################################################################



# make packages available at top-level
from .dependencies.unitsiau2015 import u,c
from .dependencies.nparray import array, linspace, arange, logspace, geomspace, invspace
from .dependencies.distl import gaussian, gaussian_around, normal, boxcar, uniform, uniform_around, histogram_from_bins, histogram_from_data, mvgaussian, mvhistogram_from_data
from .atmospheres.passbands import install_passband, uninstall_passband, uninstall_all_passbands, download_passband, list_passband_online_history, update_passband_available, update_passband, update_all_passbands, list_all_update_passbands_available, list_online_passbands, list_installed_passbands, list_passbands, list_passband_directories, get_passband
from .parameters import hierarchy, component, compute, constraint, dataset, feature, figure, solver
from .frontend.bundle import Bundle
from .backend import backends as _backends
from .solverbackends import solverbackends as _solverbackends
from . import utils as _utils

from . import dynamics as dynamics
from . import distortions as distortions
from . import algorithms as algorithms
import libphoebe

# Shortcut to building logger
def logger(*args, **kwargs):
    """
    Return a basic logger via a log file and/or terminal.

    Example 1: log only to the console, accepting levels "INFO" and above
    ```py
    logger = logger()
    ```

    Example 2: log only to the console, accepting levels "DEBUG" and above
    ```py
    logger(clevel='DEBUG')
    ```

    Example 3: log only to a file, accepting levels "DEBUG" and above
    ```py
    logger(clevel=None,filename='mylog.log')
    ```

    Example 4: log only to a file, accepting levels "INFO" and above
    ```py
    logger(clevel=None,flevel='INFO',filename='mylog.log')
    ```

    Example 5: log to the terminal (INFO and above) and file (DEBUG and above)
    ```py
    logger(filename='mylog.log')
    ```

    Arguments
    ----------
    * `clevel` (string, optional): level to be logged to the console.
        One of: "ERROR", "WARNING", "INFO", "DEBUG".
    * `flevel` (string, optional): level to be logged to the file.
        Must also provide `filename`.  One of: "ERROR", "WARNING", "INFO", "DEBUG".
    * `filename` (string, optional): path to the file to log at the `flevel` level.
    * `style` (string, optional, default='default'): style to use for logging.
        One of: "default", "minimal", "grandpa".
    """
    if mpi.within_mpirun and mpi.myrank == 0:
        # tell the workers to invoke the same logger
        mpi.comm.bcast({'worker_command': 'logger', 'args': args, 'kwargs': kwargs}, root=0)

    return _utils.get_basic_logger(*args, **kwargs)


if mpi.within_mpirun and mpi.enabled and mpi.myrank != 0:
    while True:
        packet = mpi.comm.bcast(None, root=0)

        if packet.get('worker_command', False) == 'shutdown':
            _logger.debug("rank:{}/{} message to shutdown".format(mpi.myrank, mpi.nprocs))
            exit()

        if packet.get('worker_command', False) == 'release':
            _logger.debug("rank:{}/{} message to release".format(mpi.myrank, mpi.nprocs))
            break

        elif packet.get('worker_command', False) == 'logger':
            _logger.debug("rank:{}/{} message to invoke logger".format(mpi.myrank, mpi.nprocs))
            logger(*packet['args'], **packet['kwargs'])

        elif hasattr(_backends, packet.get('backend', False)):
            backend = getattr(_backends, packet.pop('backend'))()
            backend._run_worker(packet)

        elif hasattr(_solverbackends, packet.get('backend', False)):
            backend = getattr(_solverbackends, packet.pop('backend'))()
            backend._run_worker(packet)

        else:
            raise ValueError("could not recognize packet: {}".format(packet))





# Shortcuts to bundle classmethods
def open(*args, **kwargs):
    return Bundle.open(*args, **kwargs)

open.__doc__ = Bundle.open.__doc__

def load(*args, **kwargs):
    return Bundle.open(*args, **kwargs)

load.__doc__ = Bundle.open.__doc__

def from_legacy(*args, **kwargs):
    return Bundle.from_legacy(*args, **kwargs)

from_legacy.__doc__ = Bundle.from_legacy.__doc__

def from_server(*args, **kwargs):
    return Bundle.from_server(*args, **kwargs)

from_server.__doc__ = Bundle.from_server.__doc__

def default_star(*args, **kwargs):
    return Bundle.default_star(*args, **kwargs)

default_star.__doc__ = Bundle.default_star.__doc__

def default_binary(*args, **kwargs):
    return Bundle.default_binary(*args, **kwargs)

default_binary.__doc__ = Bundle.default_binary.__doc__

def default_contact_binary(*args, **kwargs):
    return Bundle.default_contact_binary(*args, **kwargs)

default_contact_binary.__doc__ = Bundle.default_contact_binary.__doc__

def default_triple(*args, **kwargs):
    return Bundle.default_triple(*args, **kwargs)

default_triple.__doc__ = Bundle.default_triple.__doc__

# Shortcuts to settings
def reset_settings():
    """
    Reset all configuration settings (interactivity, etc) but NOT MPI settings.

    See also:
    * <phoebe.interactive_on>
    * <phoebe.interactive_off>
    * <phoebe.interactive_constraints_on>
    * <phoebe.interactive_constraints_off>
    * <phoebe.interactive_checks_on>
    * <phoebe.interactive_checks_off>
    """
    conf.reset()

def interactive_on():
    """
    Turn on both interactive constraints and interactive checks

    See also:
    * <phoebe.interactive_off>
    * <phoebe.interactive_constraints_on>
    * <phoebe.interactive_checks_on>
    """
    conf.interactive_on()

def interactive_off():
    """
    **USE WITH CAUTION**

    Turn off both interactive constraints and interactive checks

    See also:
    * <phoebe.interactive_on>
    * <phoebe.interactive_constraints_off>
    * <phoebe.interactive_checks_off>
    """
    conf.interactive_off()

def interactive_constraints_on():
    """
    Turn interactive constraints on.  When enabled, PHOEBE will update all
    constraints whenever a <phoebe.parameters.Parameter> value is changed.
    Although this adds to the run-time, it ensures that all values are updated
    when accessed.

    By default, interactive constraints are always on unless disabled.

    See also:
    * <phoebe.interactive_constraints_off>
    """
    conf.interactive_constraints_on()

def interactive_constraints_off():
    """
    **USE WITH CAUTION**

    Turn interactive constraints off.  When disabled, PHOEBE will **NOT** update
    constraints whenever a <phoebe.parameters.Parameter> value is changed, but
    will instead wait until needed (for example, by
    <phoebe.frontend.bundle.Bundle.run_compute>).  Accessing/printing the value
    of a constrained Parameter, may be out-of-date when interactive constraints
    is off.

    By default, interactive constraints are always on unless disabled.

    To update constraints manually, you can call
    <phoebe.frontend.bundle.Bundle.run_delayed_constraints>.

    See also:
    * <phoebe.interactive_constraints_on>
    """
    conf.interactive_constraints_off()

def interactive_checks_on():
    """
    Turn interactive checks on.  When enabled, PHOEBE will run system checks
    (<phoebe.frontend.bundle.Bundle.run_checks>) after any
    <phoebe.parameters.Parameter> value is changed and will log any issues
    to the logger as a warning.  In order to see these messages, you must
    have a logger enabled with at least the "WARNING" level (see <phoebe.logger>).

    Whether interactive checks is on or off, system checks will be run when
    calling <phoebe.frontend.bundle.Bundle.run_compute> and will raise
    an error if failing.

    By default, interactive checks is ON if running PHOEBE in an interactive
    console (or Jupyter notebook), but OFF if running in a script (to save
    time but also save confusing logger messages).

    See also:
    * <phoebe.interactive_checks_off>
    """
    conf.interactive_checks_on()

def interactive_checks_off():
    """
    Turn interactive checks off.  When disabled, PHOEBE will **NOT** run system checks
    (<phoebe.frontend.bundle.Bundle.run_checks>) after any
    <phoebe.parameters.Parameter> value is changed and will **NOT** log any issues
    to the logger as a warning.

    Whether interactive checks is on or off, system checks will be run when
    calling <phoebe.frontend.bundle.Bundle.run_compute> and will raise
    an error if failing.

    To manually run system checks at any time, you can call
    <phoebe.frontend.bundle.Bundle.run_checks>.

    By default, interactive checks is ON if running PHOEBE in an interactive
    console (or Jupyter notebook), but OFF if running in a script (to save
    time but also save confusing logger messages).

    See also:
    * <phoebe.interactive_checks_on>
    """
    conf.interactive_checks_off()

def devel_on():
    conf.devel_on()

def devel_off():
    conf.devel_off()

def set_download_passband_defaults(**kwargs):
    """
    Set default options to use for <phoebe.atmospheres.passbands.download_passband>.

    These can also be set at import time via the following environment variables:
    * PHOEBE_DOWNLOAD_PASSBAND_DEFAULTS_CONTENT (defaults to 'all')
    * PHOEBE_DOWNLOAD_PASSBAND_DEFAULTS_GZIPPED (defaults to FALSE)

    See also:
    * <phoebe.get_download_passband_defaults>
    * <phoebe.atmospheres.passbands.download_passband>
    * <phoebe.atmospheres.passbands.update_passband>
    * <phoebe.atmospheres.passbands.get_passband>

    Arguments
    ------------
    * `content` (string or list, optional): override the current value for
        `content` in <phoebe.get_download_passband_defaults>.
    * `gzipped` (bool, optional): override the current value for `gzipped`
        in <phoebe.get_download_passband_defaults>.

    """
    conf.set_download_passband_defaults(**kwargs)

def get_download_passband_defaults():
    """
    Access default options to use for <phoebe.atmospheres.passbands.download_passband>.

    See also:
    * <phoebe.set_download_passband_defaults>
    * <phoebe.atmospheres.passbands.download_passband>
    * <phoebe.atmospheres.passbands.update_passband>
    * <phoebe.atmospheres.passbands.get_passband>

    Returns
    ---------
    * dictionary of defaults, including `content` and `gzipped`.
    """
    return conf.get_download_passband_defaults()

def update_passband_ignore_version_on():
    """
    Turn ingoring passband versions when checking for necessary updates on.
    <phoebe.frontend.bundle.Bundle.run_checks_compute> checks to see if any
    additional content is required from the used passbands.  If so, these will
    be queried from the online tables if the timestamps match.  Otherwise, an
    error will be raised requiring manually calling <phoebe.atmospheres.passbands.update_passband>.
    By enabling this, this version conflict will be ignored, preventing the need
    to manually update the passbands.

    This can also be set at import time via the following environment variables:
    * PHOEBE_UPDATE_PASSBAND_IGNORE_VERSION (defaults to FALSE)

    See also:
    * <phoebe.update_passband_ignore_version_off>

    """
    conf.update_passband_ignore_version_on()

def update_passband_ignore_version_off():
    """
    Turn ingoring passband versions when checking for necessary updates off.
    <phoebe.frontend.bundle.Bundle.run_checks_compute> checks to see if any
    additional content is required from the used passbands.  If so, these will
    be queried from the online tables if the timestamps match.  Otherwise, an
    error will be raised requiring manually calling <phoebe.atmospheres.passbands.update_passband>.

    This can also be set at import time via the following environment variables:
    * PHOEBE_UPDATE_PASSBAND_IGNORE_VERSION (defaults to FALSE)

    See also:
    * <phoebe.update_passband_ignore_version_on>
    """
    conf.update_passband_ignore_version_on()

# Shortcuts to MPI options
def mpi_on(nprocs=None):
    """
    ENABLE PHOEBE to use MPI (parallelization).

    Default case:
    * If PHOEBE is run within an mpirun environment, MPI is ENABLED by default.
    * If PHOEBE is not run within an mpirun environment, MPI is DISABLED by default.

    When MPI is enabled, PHOEBE will do the following:
    * if within mpirun: uses PHOEBE's built-in per-dataset or per-time
        parallelization for <phoebe.frontend.bundle.Bundle.run_compute>
        and per-model parallelization when possible for
        <phoebe.frontend.bundle.Bundle.run_solver>.
    * if not within mpirun (ie. in a serial python environment): will spawn a
        separate thread at <phoebe.frontend.bundle.Bundle.run_compute>
        and <phoebe.frontend.bundle.Bundle.run_solver>,
        using `nprocs` processors.  This separate thread will be detached
        from the main thread if sending `detach=True` to
        <phoebe.frontend.bundle.Bundle.run_compute> or
        <phoebe.frontend.bundle.Bundle.run_solver>.

    See also:
    * <phoebe.mpi_off>

    Arguments
    ----------
    * `nprocs` (int, optional): number of processors.  Only applicable if **NOT**
        within mpirun (see above).
    """
    mpi.on(nprocs=nprocs)

def mpi_off():
    """
    Run PHOEBE in Serial Mode.

    Default case:
    * If PHOEBE is run within an mpirun environment, MPI is ENABLED by default.
    * If PHOEBE is not run within an mpirun environment, MPI is DISABLED by default.

    When MPI is disabled, PHOEBE will do the following:
    * if within mpirun: PHOEBE will run equally on all processors.  The user can
        customize parallelization with access to `phoebe.mpi.nprocs`,
        `phoebe.mpi.myrank`.
    * if not within mpirun (ie. in a serial python environment): PHOEBE will
        run on a single processor in serial-mode.  Compute jobs can still
        be detached from the main thread by sending `detach=True` to
        <phoebe.frontend.bundle.Bundle.run_compute> or
        <phoebe.frontend.bundle.Bundle.run_solver> but will still run
        on a single processor.

    See also:
    * <phoebe.mpi_on>
    """
    mpi.off()

def multiprocessing_on():
    """
    **NEW IN 2.3.26**

    Enable multiprocessing to use all CPUs available (this is the state by default).
    MPI will always take preference over multiprocessing.  See <phoebe.mpi_on>
    and <phoebe.mpi_off>.

    Multiprocessing is used by
    <phoebe.frontend.bundle.Bundle.run_solver> (for some solvers) and
    <phoebe.frontend.bundle.Bundle.run_compute> when `sample_from` is used.

    See also:
    * <phoebe.multiprocessing_off>
    * <phoebe.multiprocessing_get_nprocs>
    * <phoebe.multiprocessing_set_nprocs>
    """
    conf.multiprocessing_on()

def multiprocessing_off():
    """
    **NEW IN 2.3.26**

    Disable multiprocessing and force serial mode (if MPI is also off: see
    <phoebe.mpi_on> and <phoebe.mpi_off>).

    See also:
    * <phoebe.multiprocessing_on>
    * <phoebe.multiprocessing_get_nprocs>
    * <phoebe.multiprocessing_set_nprocs>
    """
    conf.multiprocessing_off()

def multiprocessing_get_nprocs():
    """
    **NEW IN 2.3.26**

    Get the number of processors used within multiprocessing.

    MPI will always take preference over multiprocessing.  See <phoebe.mpi_on>
    and <phoebe.mpi_off>.

    Multiprocessing is used by
    <phoebe.frontend.bundle.Bundle.run_solver> (for some solvers) and
    <phoebe.frontend.bundle.Bundle.run_compute> when `sample_from` is used.

    See also:
    * <phoebe.multiprocessing_on>
    * <phoebe.multiprocessing_off>
    * <phoebe.multiprocessing_set_nprocs>
    """
    return conf.multiprocessing_nprocs

def multiprocessing_set_nprocs(nprocs):
    """
    **NEW IN 2.3.26**

    Set a custom number of processors to use within multiprocessing.
    MPI will always take preference over multiprocessing.  See <phoebe.mpi_on>
    and <phoebe.mpi_off>.

    Multiprocessing is used by
    <phoebe.frontend.bundle.Bundle.run_solver> (for some solvers) and
    <phoebe.frontend.bundle.Bundle.run_compute> when `sample_from` is used.

    See also:
    * <phoebe.multiprocessing_on>
    * <phoebe.multiprocessing_off>
    * <phoebe.multiprocessing_get_nprocs>
    """
    conf.multiprocessing_set_nprocs(nprocs)

def progressbars_on():
    """
    Enable progressbars. Progressbars require `tqdm` to be installed
    (will silently ignore if not installed).

    See also:
    * <phoebe.progressbars_off>
    """
    conf.progressbars_on()

def progressbars_off():
    """
    Disable progressbars. Progressbars require `tqdm` to be installed
    (will silently ignore if not installed).

    See also:
    * <phoebe.progressbars_on>
    """
    conf.progressbars_off()

# let's use magic to shutdown the workers when the user-script is complete
atexit.register(mpi.shutdown_workers)

# edit API docs for imported functions

def strip_docstring_refs(matchobj):
    text = matchobj.group(0)
    path = text[1:-1]
    return path

def add_nparray_docstring(obj):

    docsprefix = """This is an included dependency from [nparray 1.2.0](https://nparray.readthedocs.io/en/1.2.0/).\n\n===============================================================\n\n"""

    docstring = docsprefix + "\n".join([l.lstrip() for l in obj.__doc__.split("\n")])
    docstring = re.sub(r"(?P<name>\<[0-9a-zA-Z_\.]*\>)", strip_docstring_refs, docstring)

    obj.__doc__ = docstring

add_nparray_docstring(array)
add_nparray_docstring(linspace)
add_nparray_docstring(arange)
add_nparray_docstring(logspace)
add_nparray_docstring(geomspace)
add_nparray_docstring(invspace)


def add_distl_docstring(obj):
    docsprefix = """This is an included dependency from [distl](https://distl.readthedocs.io).\n\n===============================================================\n\n"""

    docstring = docsprefix + "\n".join([l.lstrip() for l in obj.__doc__.split("\n")])
    docstring = re.sub(r"(?P<name>\<[0-9a-zA-Z_\.]*\>)", strip_docstring_refs, docstring)

    obj.__doc__ = docstring

add_distl_docstring(uniform)
add_distl_docstring(boxcar)
# add_distl_docstring(delta)
add_distl_docstring(gaussian)
add_distl_docstring(normal)
add_distl_docstring(histogram_from_bins)
add_distl_docstring(histogram_from_data)
add_distl_docstring(mvgaussian)
add_distl_docstring(mvhistogram_from_data)
add_distl_docstring(uniform_around)
add_distl_docstring(gaussian_around)


# expose available "kinds" per-context
def _get_phoebe_funcs(module, devel=False):
    ignore = ['_empty_array', 'deepcopy', 'fnmatch',
              'download_passband', 'list_installed_passbands', 'list_online_passbands', 'list_passbands', 'parameter_from_json', 'parse_json',
              'send_if_client', 'update_if_client',
              '_add_component', '_add_dataset', '_label_units_lims', '_run_compute',
              'phase_mask_inds']

    if not devel:
        ignore += ['pulsation']
        ignore += ['photodynam']

    mod_split = module.__name__.split('.')
    if mod_split[-1] in ['figure'] or (mod_split[-1] in ['solver'] and 'figure' not in mod_split):
        ret = []
        for sub_module in _inspect.getmembers(module):
            if _inspect.ismodule(sub_module[1]):
                ret += [sub_module[0]+"."+o for o in _get_phoebe_funcs(sub_module[1], devel=devel)]
        return ret

    return [o[0] for o in _inspect.getmembers(module) if _inspect.isfunction(o[1]) and o[0] not in ignore and o[0][0] != '_']


def list_available_components(devel=False):
    """
    List all available 'kinds' for component from <phoebe.parameters.component>.

    See also:
    * <phoebe.list_available_features>
    * <phoebe.list_available_datasets>
    * <phoebe.list_available_computes>
    * <phoebe.list_available_solvers>

    Arguments
    -----------
    * `devel` (bool, default, optional=False): whether to include development-only
        kinds.  See <phoebe.devel_on>.

    Returns
    ---------
    * (list of strings)
    """
    return _get_phoebe_funcs(component, devel=devel)

def list_available_features(devel=False):
    """
    List all available 'kinds' for feature from <phoebe.parameters.feature>.

    See also:
    * <phoebe.list_available_components>
    * <phoebe.list_available_datasets>
    * <phoebe.list_available_computes>
    * <phoebe.list_available_solvers>

    Arguments
    -----------
    * `devel` (bool, default, optional=False): whether to include development-only
        kinds.  See <phoebe.devel_on>.

    Returns
    ---------
    * (list of strings)
    """
    return _get_phoebe_funcs(feature, devel=devel)

def list_available_datasets(devel=False):
    """
    List all available 'kinds' for dataset from <phoebe.parameters.dataset>.

    See also:
    * <phoebe.list_available_components>
    * <phoebe.list_available_features>
    * <phoebe.list_available_computes>
    * <phoebe.list_available_solvers>

    Arguments
    -----------
    * `devel` (bool, default, optional=False): whether to include development-only
        kinds.  See <phoebe.devel_on>.

    Returns
    ---------
    * (list of strings)
    """
    return  _get_phoebe_funcs(dataset, devel=devel)

def list_available_figures(devel=False):
    """
    List all available 'kinds' for figure from <phoebe.parameters.figure>.

    See also:
    * <phoebe.list_available_components>
    * <phoebe.list_available_features>
    * <phoebe.list_available_computes>
    * <phoebe.list_available_solvers>

    Arguments
    -----------
    * `devel` (bool, default, optional=False): whether to include development-only
        kinds.  See <phoebe.devel_on>.

    Returns
    ---------
    * (list of strings)
    """
    return  _get_phoebe_funcs(figure, devel=devel)

def list_available_computes(devel=False):
    """
    List all available 'kinds' for compute from <phoebe.parameters.compute>.

    See also:
    * <phoebe.list_available_components>
    * <phoebe.list_available_features>
    * <phoebe.list_available_datasets>
    * <phoebe.list_available_solvers>

    Arguments
    -----------
    * `devel` (bool, default, optional=False): whether to include development-only
        kinds.  See <phoebe.devel_on>.

    Returns
    ---------
    * (list of strings)
    """
    return _get_phoebe_funcs(compute, devel=devel)

def list_available_solvers(devel=False):
    """
    List all available 'kinds' for solver from <phoebe.parameters.solver>.

    See also:
    * <phoebe.list_available_components>
    * <phoebe.list_available_features>
    * <phoebe.list_available_datasets>
    * <phoebe.list_available_computes>

    Arguments
    -----------
    * `devel` (bool, default, optional=False): whether to include development-only
        kinds.  See <phoebe.devel_on>.

    Returns
    ---------
    * (list of strings)
    """
    return _get_phoebe_funcs(solver, devel=devel)

if _env_variable_bool('PHOEBE_ENABLE_ONLINE_PASSBANDS', 'TRUE'):
    for pb in list_all_update_passbands_available():
        msg = 'passband "{}" has a newer version available.  Run phoebe.list_passband_online_history("{}") to get a list of available changes and phoebe.update_passband("{}") or phoebe.update_all_passbands() to update.'.format(pb, pb, pb)
        # NOTE: we'll print since the logger hasn't been initialized yet.
        print('PHOEBE: {}'.format(msg))

# delete things we don't want exposed to the user at the top-level
# NOTE: we need _sys for reset_settings, that's why its __sys
del atexit
try:
    del matplotlib
except:
    pass
try:
    del mpi4py
except:
    pass

del logging
del Settings
del MPI

del re
del strip_docstring_refs
del add_nparray_docstring
del add_distl_docstring




from phoebe.parameters import *

def _to_component(obj, allow_hierarchy=True):
    """
    takes either a string, ParameterSet, Parameter, or the string representation (component or nested hierarchy)
    """

    def _to_str(obj):
        return "{}:{}".format(obj.kind, obj.component)

    if isinstance(obj, str):
        # TODO: check to see if valid?, use allow_hierarchy
        # TODO: when passed labels this is going to give the wrong thing, but that might be fixable in the HierarchyParameter set_value check
        #~ raise NotImplementedError # need to get object from bundle and pass to _to_str
        return obj
    elif isinstance(obj, ParameterSet):
        # TODO: be smarter about this and don't assume only 1 will be returned
        if 'repr' in obj.to_flat_dict().keys():
            return obj.get_value(qualifier='repr')

        else:
            # make sure we only have things in the 'component' context
            obj = obj.filter(context='component')
            return _to_str(obj)

    elif isinstance(obj, Parameter):
        if obj.qualifier == 'repr':
            return obj.get_value()
        else:
            return _to_str(obj)

    else:
        raise NotImplementedError("could not parse {}".format(obj))


def binaryorbit(orbit, comp1, comp2, envelope=None):
    """
    Create the string representation of a hierarchy containing a binary orbit
    with two components.

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.set_hierarchy>.  If attaching through
    <phoebe.frontend.bundle.Bundle.set_hierarchy>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.

    Arguments
    ----------
    * `comp1` (string or <phoebe.parameters.Parameter> or
        <phoebe.parameters.ParameterSet): primary component.
    * `comp2` (string or <phoebe.parameters.Parameter> or
        <phoebe.parameters.ParameterSet): secondary component.
    * `env` (string or <phoebe.parameters.Parameter> or
        <phoebe.parameters.ParameterSet, optional, default=None): envelope
        component.  If provided, this will create a contact binary system,
        otherwise a detached system will be created.

    Returns
    --------
    * (str): the string representation of the hierarchy, ready to be sent to
        <phoebe.frontend.bundle.Bundle.set_hierarchy>.
    """

    if envelope:
        return '{}({}, {}, {})'.format(_to_component(orbit, False), _to_component(comp1), _to_component(comp2), _to_component(envelope, False))
    else:
        return '{}({}, {})'.format(_to_component(orbit, False), _to_component(comp1), _to_component(comp2))

def component(*args):
    """
    Create the string representation of a hierarchy that groups multiple objects
    without a parent orbit (ie. a single star or a  disk around a planet).

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.set_hierarchy>.  If attaching through
    <phoebe.frontend.bundle.Bundle.set_hierarchy>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.

    Arguments
    ----------
    * `*args` (string or <phoebe.parameters.Parameter> or
        <phoebe.parameters.ParameterSet): any number of individual components
        can be passed to this function.

    Returns
    --------
    * (str): the string representation of the hierarchy, ready to be sent to
        <phoebe.frontend.bundle.Bundle.set_hierarchy>.
    """

    return '{}'.format(', '.join([_to_component(arg) for arg in args]))

def blank(*args):
    """
    Create the string representation of a blank hierarchy.

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.set_hierarchy>.  If attaching through
    <phoebe.frontend.bundle.Bundle.set_hierarchy>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.

    Arguments
    ----------
    * `*args`: IGNORED

    Returns
    --------
    * (str): the string representation of the hierarchy, ready to be sent to
        <phoebe.frontend.bundle.Bundle.set_hierarchy>.
    """
    return ''
